"""This module implements the correlation estimators and a way to symbolically
represent paircounts (e.g. data-data counts are represented by :obj:`CtsDD`).

The latter is used in :obj:`~yaw.correlation.CorrFunc` to determine which
correlation estimator can be computed for a possibly incomplete set of pair
counts (e.g. only data-data and random-random). Their key property is that they
can be compared, e.g.

>>> CtsDR() == CtsRR()  # random-data is random-random?
False

A special case is the :obj:`DavisPeebles` Estimator, which is the ratio
:math:`DD/DR-1`. In the case of a crosscorrelation it is irrelevant which of the
samples provides a random sample. Therefore, there is the special class
:obj:`CtsMix`, with the property:

>>> CtsMix() == CtsDR()
True
>>> CtsMix() == CtsRD()
True
"""

from __future__ import annotations

import logging
from abc import ABC, abstractclassmethod, abstractproperty
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

__all__ = [
    "CorrelationEstimator",
    "PeeblesHauser",
    "DavisPeebles",
    "Hamilton",
    "LandySzalay",
]


logger = logging.getLogger(__name__)


class EstimatorError(Exception):
    pass


class Cts(ABC):
    """Base class to symbolically represent pair counts."""

    @abstractproperty
    def _hash(self) -> int:
        """Used for comparison.

        Equivalent pair counts types should have the same hash value, see
        :obj:`CtsMix`."""
        pass

    @abstractproperty
    def _str(self) -> str:
        pass

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}>"

    def __str__(self) -> str:
        return self._str

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Cts):
            return False
        var1 = set(self._str.split("_"))
        var2 = set(other._str.split("_"))
        return not var1.isdisjoint(var2)


class CtsDD(Cts):
    """Symbolic representation of data-data paircounts."""

    _str = "dd"
    _hash = 1


class CtsDR(Cts):
    """Symbolic representation of data-random paircounts."""

    _str = "dr"
    _hash = 2


class CtsRD(Cts):
    """Symbolic representation of random-data paircounts."""

    _str = "rd"
    _hash = 2


class CtsMix(Cts):
    """Symbolic representation of either data-random or random-data paircounts.

    >>> CtsDR() == CtsMix()
    True
    >>> CtsRD() == CtsMix()
    True
    """

    _str = "dr_rd"
    _hash = 2


class CtsRR(Cts):
    """Symbolic representation of random-random paircounts."""

    _str = "rr"
    _hash = 3


def cts_from_code(code: str) -> Cts:
    """Instantiate the correct :obj:`Cts` subclass from a string.

    Valid options are ``dd``, ``dr``, ``rd``, ``rr``, e.g.:

    >>> cts_from_code("dr")
    <CtsDR>
    """
    codes = dict(dd=CtsDD, dr=CtsDR, rd=CtsRD, rr=CtsRR)
    if code not in codes:
        raise ValueError(f"unknown pair counts '{code}'")
    return codes[code]()


class CorrelationEstimator(ABC):
    name: str
    """Full name of the estimator."""
    short: str
    """Get a short form representation of the estimator name."""
    requires: list[Cts]
    """Get a symbolic list of pair counts required to evaluate the estimator.
    """
    optional: list[Cts]
    """Get a symbolic list of optional pair counts that may be used to evaluate
    the estimator.
    """

    variants: list[CorrelationEstimator] = []
    """List of all implemented correlation estimators classes."""

    def __init_subclass__(cls, **kwargs):
        """Add all subclusses to the :obj:`variants` attribute list."""
        super().__init_subclass__(**kwargs)
        cls.variants.append(cls)

    @classmethod
    def _warn_enum_zero(cls, counts: NDArray):
        """Raise a warning if any value in the expression enumerator is zero"""
        if np.any(counts == 0.0):
            logger.warn(f"estimator {cls.short} encontered zeros in enumerator")

    @abstractclassmethod
    def eval(
        cls, *, dd: NDArray, dr: NDArray, rr: NDArray, rd: NDArray | None = None
    ) -> NDArray:
        """Method that implements the estimator.

        Should call :meth:`_warn_enum_zero` to raise warnings on zero-division.
        """
        NotImplemented


class PeeblesHauser(CorrelationEstimator):
    """Implementation of the Peebles-Hauser correlation estimator
    :math:`\\frac{DD}{RR} - 1`.
    """

    name = "PeeblesHauser"
    short = "PH"
    requires = [CtsDD(), CtsRR()]
    optional = []

    @classmethod
    def eval(cls, *, dd: NDArray, rr: NDArray, **kwargs) -> NDArray:
        """Evaluate the estimator with the given pair counts.

        Args:
            dd (:obj:`NDArray`):
                Data-data pair counts (normalised).
            rr (:obj:`NDArray`):
                Random-random pair counts (normalised).

        Returns:
            :obj:`NDArray`
        """
        cls._warn_enum_zero(rr)
        return dd / rr - 1.0


class DavisPeebles(CorrelationEstimator):
    """Implementation of the Davis-Peebles correlation estimator
    :math:`\\frac{DD}{DR} - 1`.

    .. Note::
        Accepts both :math:`DR` and :math:`RD` for the denominator.
    """

    name = "DavisPeebles"
    short = "DP"
    requires = [CtsDD(), CtsMix()]
    optional = []

    @classmethod
    def eval(cls, *, dd: NDArray, dr_rd: NDArray, **kwargs) -> NDArray:
        """Evaluate the estimator with the given pair counts.

        Args:
            dd (:obj:`NDArray`):
                Data-data pair counts (normalised).
            dr_rd (:obj:`NDArray`):
                Either data-random or random-data pair counts (normalised).

        Returns:
            :obj:`NDArray`
        """
        cls._warn_enum_zero(dr_rd)
        return dd / dr_rd - 1.0


class Hamilton(CorrelationEstimator):
    """Implementation of the Hamilton correlation estimator
    :math:`\\frac{DD \\times RR}{DR \\times RD} - 1`.
    """

    name = "Hamilton"
    short = "HM"
    requires = [CtsDD(), CtsDR(), CtsRR()]
    optional = [CtsRD()]

    @classmethod
    def eval(
        cls, *, dd: NDArray, dr: NDArray, rr: NDArray, rd: NDArray | None = None
    ) -> NDArray:
        """Evaluate the estimator with the given pair counts.

        Args:
            dd (:obj:`NDArray`):
                Data-data pair counts (normalised).
            dr (:obj:`NDArray`):
                Data-random pair counts (normalised).
            rd (:obj:`NDArray`, optional):
                Random-data pair counts (normalised). If not provided, use
                ``dr`` instead.
            rr (:obj:`NDArray`):
                Random-random pair counts (normalised).

        Returns:
            :obj:`NDArray`
        """
        rd = dr if rd is None else rd
        enum = dr * rd
        cls._warn_enum_zero(enum)
        return (dd * rr) / enum - 1.0


class LandySzalay(CorrelationEstimator):
    """Implementation of the Landy-Szalay correlation estimator
    :math:`\\frac{DD - (DR + RD)}{RR} + 1`.
    """

    name = "LandySzalay"
    short = "LS"
    requires = [CtsDD(), CtsDR(), CtsRR()]
    optional = [CtsRD()]

    @classmethod
    def eval(
        cls, *, dd: NDArray, dr: NDArray, rr: NDArray, rd: NDArray | None = None
    ) -> NDArray:
        """Evaluate the estimator with the given pair counts.

        Args:
            dd (:obj:`NDArray`):
                Data-data pair counts (normalised).
            dr (:obj:`NDArray`):
                Data-random pair counts (normalised).
            rd (:obj:`NDArray`, optional):
                Random-data pair counts (normalised). If not provided, use
                ``dr`` instead.
            rr (:obj:`NDArray`):
                Random-random pair counts (normalised).

        Returns:
            :obj:`NDArray`
        """
        rd = dr if rd is None else rd
        cls._warn_enum_zero(rr)
        return (dd - (dr + rd) + rr) / rr



"""This module implements a class that generates the value options for certain
configuration parameters. An instance of this class can be accessed directly as
:obj:`yaw.config.OPTIONS`.
"""

from __future__ import annotations

__all__ = ["OPTIONS"]


class Options:
    @property
    def backend(self) -> tuple[str]:
        """Lists the names of the currently available backends for correlation
        measurements."""
        from yaw.catalogs import BaseCatalog

        return tuple(sorted(BaseCatalog._backends.keys()))

    @property
    def binning(self) -> tuple[str]:
        """Lists the currently implemented methods to generate redshift bins
        with the :obj:`~yaw.config.AutoBinningConfig` class.

        .. rubric:: Values

        ``comoving``: Generate a binning with equal width in radial comoving
        distance.

        ``linear``: Generate a binning with equal width in redshift.

        ``logspace``: Generate a binning with equal width in logarithmic
        redshift :math:`\\log(1+z)`.
        """
        return ("comoving", "linear", "logspace")

    @property
    def cosmology(self) -> tuple[str]:
        """Lists the availble named cosmologies in :obj:`astropy`.

        On top of these comological models, custom cosmologies can be defined by
        subclassing :obj:`yaw.core.cosmology.CustomCosmology`.
        """
        from astropy.cosmology import available

        return available

    @property
    def kind(self) -> tuple[str]:
        """Lists the currently implemented methods for covariance calculation.

        .. rubric:: Values

        ``full``: Compute all matrix elements of the covariance.

        ``diag``: Compute only the main diagonal and the primary off-diagonals
        of the covariance matrix. This option applies, if the covariance is
        computed from a concatenated set of data samples, which have a
        crosscorrelation of interest. For example if concatenating the samples
        obtained from multiple redshift bins, the primary off-diagonals contain
        the covariance at the same redshift between different bins.

        ``var``: Compute the variance, i.e. only the diagonal elements.
        """
        return ("full", "diag", "var")

    @property
    def merge(self) -> tuple[str]:
        """Lists the available modes to merge correlation measurements.

        .. rubric:: Values

        ``patches``: Merge measurements by concatenating spatial patches, i.e.
        extending the area over which measurements are taken. This may miss out
        some correlation signal between the two measurements.

        ``redshift``: Merge measurements by concatenating redshift bins, i.e.
        extending the redshift range.
        """
        return ("patches", "redshift")

    @property
    def method(self) -> tuple[str]:
        """Lists the currently implemented methods for spatial resampling.

        Resampling uses the spatial patches to get uncertainty estimates for the
        correlation function measurements.

        .. rubric:: Values

        ``jackknife``: Use jackknife resampling (generate samples and leave out
        one patch at a time).

        ``bootstrap``: Use bootstrap resampling (generate samples by randomly
        drawing patches with replacement).
        """
        return ("jackknife", "bootstrap")


OPTIONS = Options()



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from numpy.typing import NDArray
from scipy.spatial import distance_matrix

from yaw.core.containers import PatchIDs
from yaw.core.coordinates import Dist3D, DistSky
from yaw.core.cosmology import r_kpc_to_angle

if TYPE_CHECKING:  # pragma: no cover
    from yaw.catalogs import BaseCatalog
    from yaw.config import Configuration

__all__ = ["PatchLinkage"]


logger = logging.getLogger(__name__)


LINK_ZMIN = 0.05
"""The reference redshift at which the maximum angular size is computed."""


class PatchLinkage:
    """Class that links patches for pair counting, serves as task generator.

    This class is useful to generate pairs of patches that need to be paired
    for pair count measurements for a given maximum angular search radius.
    Patches that are separated by more than the sum of their radii and the
    maximum search radius do not contribute any pair counts and can be
    discarded.
    """

    def __init__(self, patch_tuples: list[PatchIDs]) -> None:
        """Populate a patch linkage container.

        To create a new linkage, use the :meth:`from_setup` method.

        Args:
            patch_tuples (list[:obj:`~yaw.core.containers.PatchIDs`]):
                List of patch pairs that need to be visited to count pairs.
        """
        self.pairs = patch_tuples

    @classmethod
    def from_setup(cls, config: Configuration, catalog: BaseCatalog) -> PatchLinkage:
        """Generate a new patch linkage.

        Compute a maximum angular separation for at low redshift for the scales
        provided in the configuration. Generate a list of all patch pairs that
        are separated by less than this maximum separation (factoring in the
        size of the patches).

        Args:
            config (`~yaw.Configuration`):
                Configuration object that defines the scales and cosmology
                needed to compute the maximum angular scale.
            catalog (:obj:`BaseCatalog`):
                Catalog instance with patch centers and sizes.

        Returns:
            :obj:`PatchLinkage`
        """
        # determine the additional overlap from the spatial query
        if config.backend.crosspatch:
            # estimate maximum query radius at low, but non-zero redshift
            z_ref = max(LINK_ZMIN, config.binning.zmin)
            max_query_radius = r_kpc_to_angle(
                config.scales.as_array(), z_ref, config.cosmology
            ).max()
        else:
            max_query_radius = 0.0  # only relevant for cross-patch
        max_query_radius = DistSky(max_query_radius)

        logger.debug(f"computing patch linkage with {max_query_radius=:.3e}")
        centers_3d = catalog.centers.to_3d().values
        radii = catalog.radii.values
        # compute distance between all patch centers
        dist_mat_3d = Dist3D(distance_matrix(centers_3d, centers_3d))
        # compare minimum separation required for patchs to not overlap
        size_sum = DistSky(np.add.outer(radii, radii))

        # check which patches overlap when factoring in the query radius
        overlaps = dist_mat_3d.to_sky() < (size_sum + max_query_radius)
        patch_pairs = []
        for id1, overlap in enumerate(overlaps):
            patch_pairs.extend((id1, id2) for id2 in np.where(overlap)[0])
        logger.debug(
            f"found {len(patch_pairs)} patch links " f"for {catalog.n_patches} patches"
        )
        return cls(patch_pairs)

    def __len__(self) -> int:
        return len(self.pairs)

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"{name}(n_patches={self.n_patches}, n_pairs={len(self)})"

    @property
    def n_patches(self) -> int:
        """Get the total number of patches."""
        patches = set()
        for p1, p2 in self.pairs:
            patches.add(p1)
            patches.add(p2)
        return len(patches)

    @property
    def density(self) -> float:
        """Get ratio of the number of linked patch pairs compared to all
        possible combinations."""
        n = self.n_patches
        return len(self) / (n * n)

    def get_pairs(self, auto: bool, crosspatch: bool = True) -> list[PatchIDs]:
        """Get a list of linked patch pairs.

        Args:
            auto (:obj:`bool`):
                For autocorrelation measurements, only visit patch pairs where
                patch ID1 >= ID2.
            crosspatch (:obj:`bool`):
                If false, ignore all cross-patch pair counts and link patches
                just with themselves.

        Returns:
            :obj:`list[PatchIDs]`
        """
        if crosspatch:
            if auto:
                pairs = [(i, j) for i, j in self.pairs if j >= i]
            else:
                pairs = self.pairs
        else:
            pairs = [(i, j) for i, j in self.pairs if i == j]
        return pairs

    @staticmethod
    def _parse_collections(
        collection1: BaseCatalog, collection2: BaseCatalog | None = None
    ) -> tuple[bool, BaseCatalog, BaseCatalog]:
        auto = collection2 is None
        if auto:
            collection2 = collection1
        return auto, collection1, collection2

    def get_matrix(
        self,
        collection1: BaseCatalog,
        collection2: BaseCatalog | None = None,
        crosspatch: bool = True,
    ) -> NDArray[np.bool_]:
        """Convert the list of linked patches to a boolean matrix indicating if
        two patches are linked.

        Depending on if one or two catalogs are provided as input, the result
        resembles a cross- or autocorrelation patch linkage. This is the only
        purpose of the inputs.

        Args:
            collection1 (:obj:`BaseCatalog`):
                First catalog for patch linkage.
            collection2 (:obj:`BaseCatalog`, optional):
                Second catalog for patch linkage. If not provided, returns a
                matrix for an autocorrelation case.
            crosspatch (:obj:`bool`):
                Link patches just with themselves and ignore cross-patch pairs.

        Returns:
            :obj:`NDArray`

        .. Warning::

            The patch centers of both catalogues must be (very close to)
            identical.
        """
        auto, collection1, collection2 = self._parse_collections(
            collection1, collection2
        )
        pairs = self.get_pairs(auto, crosspatch)
        # make a boolean matrix indicating the exisiting patch combinations
        n_patches = self.n_patches
        matrix = np.zeros((n_patches, n_patches), dtype=np.bool_)
        for pair in pairs:
            matrix[pair] = True
        return matrix

    def get_patches(
        self,
        collection1: BaseCatalog,
        collection2: BaseCatalog | None = None,
        crosspatch: bool = True,
    ) -> tuple[list[BaseCatalog], list[BaseCatalog]]:
        """Return linked pairs of patch data ready for processing.

        Instead of returning a list of patch index pairs, the actual patch data
        is returned in two aligned list, where item 1 form the first list is
        linked to item 1 of the second list. Depending on if one or two catalogs
        are provided as input, the result resembles a cross- or autocorrelation
        patch linkage.

        Args:
            collection1 (:obj:`BaseCatalog`):
                First catalog for patch linkage.
            collection2 (:obj:`BaseCatalog`, optional):
                Second catalog for patch linkage. If not provided, returns a
                matrix for an autocorrelation case.
            crosspatch (:obj:`bool`):
                Link patches just with themselves and ignore cross-patch pairs.

        Returns:
            list, list: Two lists with patch data from ``collection1`` and
            ``collection2`` (if provided, else ``collection1``) that are linked.

        .. Warning::

            The patch centers of both catalogues must be (very close to)
            identical.
        """
        auto, collection1, collection2 = self._parse_collections(
            collection1, collection2
        )
        pairs = self.get_pairs(auto, crosspatch)
        # generate the patch lists
        patches1 = []
        patches2 = []
        for id1, id2 in pairs:
            patches1.append(collection1[id1])
            patches2.append(collection2[id2])
        return patches1, patches2



"""This module implements the core library, which defines a number of utilities:

* the (abstract) API of the main data containers (:mod:`yaw.core.abc`,
  :obj:`yaw.core.SampledData`, and :obj:`yaw.core.SampledValue`),

* coordinate systems and distance computation on the sphere
  (:mod:`yaw.core.coordinates`),

* a cosmology interface and base class for custom
  cosmologies (:mod:`yaw.core.cosmology`),

* parallelisation with multiprocessing (:mod:`yaw.core.parallel`), and

* a few other convenience functions in various submodules.
"""

from yaw.core.containers import SampledData, SampledValue
from yaw.core.math import global_covariance

__all__ = ["SampledData", "SampledValue", "global_covariance"]



"""This module implements the configuration for ``yaw``, which are the free
parameters for the correlation function measurements. These parameters control

1. the scales on which correlations are measured (:obj:`ScalesConfig`),
2. the redshift binning for the redshift reconstruction
   (:obj:`AutoBinningConfig` or :obj:`ManualBinningConfig`),
3. backend related parameters (:obj:`BackendConfig`), and
4. the resampling method for error estimation (:obj:`ResamplingConfig`).

Each of these four configuration classes are implemented as immuateble
dataclasses, i.e. their values cannot be modified after creation.

For convenience, the first three configuration classes are grouped together in
the :obj:`Configuration` class, since these parameters are often needed
together. This configuration class additionally specifies the cosmological
model. For more details, refer to the :obj:`yaw.core.cosmology` module.

The recommended way to create a new configuration is thorough its constructor
methods :meth:`Configuration.create` or :meth:`Configuration.modify` to create a
new, modified configuration from an existing one. The default values for the
parameters are listed in :mod:`yaw.config.default` (also available as
``yaw.config.DEFAULT``), parameters that support a fixed set of options can be
accessed through :obj:`yaw.config.OPTIONS`, an instance of the
:obj:`~yaw.config.options.Options` generating class.
"""

from yaw.config import default as DEFAULT
from yaw.config.options import OPTIONS

# isort: split
from yaw.config.backend import BackendConfig
from yaw.config.binning import AutoBinningConfig, ManualBinningConfig
from yaw.config.config import Configuration
from yaw.config.resampling import ResamplingConfig
from yaw.config.scales import ScalesConfig

__all__ = [
    "DEFAULT",
    "BackendConfig",
    "AutoBinningConfig",
    "ManualBinningConfig",
    "Configuration",
    "OPTIONS",
    "ResamplingConfig",
    "ScalesConfig",
]



from __future__ import annotations

import itertools
import os
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from yaw.catalogs import BaseCatalog, PatchLinkage
from yaw.catalogs.scipy.patches import (
    PatchCatalog,
    assign_patches,
    create_patches,
    patch_id_from_path,
)
from yaw.config import Configuration, ResamplingConfig
from yaw.core.containers import PatchCorrelationData, PatchIDs
from yaw.core.coordinates import Coord3D, Coordinate, CoordSky, DistSky
from yaw.core.cosmology import Scale
from yaw.core.logging import TimedLog
from yaw.core.parallel import ParallelHelper
from yaw.core.utils import LimitTracker, job_progress_bar, long_num_format
from yaw.correlation.paircounts import (
    NormalisedCounts,
    PatchedCount,
    PatchedTotal,
    pack_results,
)
from yaw.redshifts import HistData

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray
    from pandas import DataFrame

    from yaw.core.cosmology import TypeCosmology

__all__ = ["ScipyCatalog"]


def _count_pairs_thread(
    patch1: PatchCatalog,
    patch2: PatchCatalog,
    scales: Sequence[Scale],
    cosmology: TypeCosmology,
    z_bins: NDArray[np.float_],
    bin1: bool = True,
    bin2: bool = False,
    dist_weight_scale: float | None = None,
    dist_weight_res: int = 50,
) -> PatchCorrelationData:
    """Implementes the pair counting between two patches.

    Bins the data as needed and builds the KDTrees for the pair finding.
    Converts the physical scales to angles for the given cosmology and redshift
    and counts the pairs. Pairs are recoreded for each set of scales and stored
    in a PatchCorrelationData object.
    """
    z_intervals = pd.IntervalIndex.from_breaks(z_bins)
    # build trees
    patch1.load(use_threads=False)
    if bin1:
        trees1 = [patch.get_tree() for _, patch in patch1.iter_bins(z_bins)]
    else:
        trees1 = itertools.repeat(patch1.get_tree())
    patch2.load(use_threads=False)
    if bin2:
        trees2 = [patch.get_tree() for _, patch in patch2.iter_bins(z_bins)]
    else:
        trees2 = itertools.repeat(patch2.get_tree())
    # count pairs, iterate through the bins and count pairs between the trees
    counts = np.empty((len(scales), len(z_intervals)))
    totals1 = np.empty(len(z_intervals))
    totals2 = np.empty(len(z_intervals))
    for i, (intv, tree1, tree2) in enumerate(zip(z_intervals, trees1, trees2)):
        # if bin1 is False and bin2 is False, these will still give different
        # counts since the angle for scales is chaning
        angles = [scale.to_radian(intv.mid, cosmology) for scale in scales]
        counts[:, i] = tree1.count(
            tree2,
            scales=angles,
            dist_weight_scale=dist_weight_scale,
            weight_res=dist_weight_res,
        )
        totals1[i] = tree1.total
        totals2[i] = tree2.total
    counts = {str(scale): count for scale, count in zip(scales, counts)}
    return PatchCorrelationData(
        patches=PatchIDs(patch1.id, patch2.id),
        totals1=totals1,
        totals2=totals2,
        counts=counts,
    )


def _histogram_thread(
    patch: PatchCatalog, z_bins: NDArray[np.float_]
) -> NDArray[np.float_]:
    """Computes a redshift histogram for a single PatchCatalog and returns the
    counts as array."""
    is_loaded = patch.is_loaded()
    patch.load()
    counts, _ = np.histogram(patch.redshifts, z_bins, weights=patch.weights)
    if not is_loaded:
        patch.unload()
    return counts


class ScipyCatalog(BaseCatalog):
    """An implementation of the :obj:`BaseCatalog` using a wrapper around
    :obj:`scipy.spatial.cKDTree` for the pair counting, which is implemented in
    :obj:`yaw.catalogs.scipy.kdtree`. Fully supports caching.

    .. Note::

        This is currently the default backend and has the best support and
        performance. Currently, trees cannot be shared across the
        multiprocessing interface and must be rebuilt every time a patch is
        used for pair counting again.
    """

    def __init__(
        self,
        data: DataFrame,
        ra_name: str,
        dec_name: str,
        *,
        patch_name: str | None = None,
        patch_centers: BaseCatalog | Coordinate | None = None,
        n_patches: int | None = None,
        redshift_name: str | None = None,
        weight_name: str | None = None,
        cache_directory: str | None = None,
        progress: bool = True,
    ) -> None:
        if len(data) == 0:
            raise ValueError("data catalog is empty")
        # check if the columns exist
        renames = {ra_name: "ra", dec_name: "dec"}
        if redshift_name is not None:
            renames[redshift_name] = "redshift"
        if weight_name is not None:
            renames[weight_name] = "weights"
        for col_name, kind in renames.items():
            if col_name not in data:
                raise KeyError(f"column {kind}='{col_name}' not found in data")

        # check if patches should be written and unloaded from memory
        unload = cache_directory is not None
        if patch_name is not None:
            patch_mode = "dividing"
        else:
            if n_patches is not None:
                patch_mode = "creating"
            elif patch_centers is not None:
                patch_mode = "applying"
            else:
                raise ValueError(
                    "either of 'patch_name', 'patch_centers', or 'n_patches' "
                    "must be provided"
                )
        if unload:
            if not os.path.exists(cache_directory):
                raise FileNotFoundError(
                    f"patch directory does not exist: '{cache_directory}'"
                )
            self._logger.debug(f"using cache directory '{cache_directory}'")

        # create new patches
        if patch_mode != "dividing":
            position = CoordSky.from_array(
                np.deg2rad(data[[ra_name, dec_name]].to_numpy())
            )
            if patch_mode == "creating":
                patch_centers, patch_ids = create_patches(
                    position=position, n_patches=n_patches
                )
                log_msg = f"creating {n_patches} patches"
            else:
                if isinstance(patch_centers, BaseCatalog):
                    patch_centers = patch_centers.centers.to_3d()
                patch_ids = assign_patches(centers=patch_centers, position=position)
                n_patches = len(patch_centers)
                log_msg = f"applying {n_patches} patches from external data"
            patch_name = "patch"  # the default name
            data[patch_name] = patch_ids
            centers = {pid: pos for pid, pos in enumerate(patch_centers)}
        else:
            n_patches = len(data[patch_name].unique())
            log_msg = f"dividing data into {n_patches} predefined patches"
            centers = dict()  # this can be empty
        self._logger.debug(log_msg)

        # run groupby first to avoid any intermediate copies of full data
        n_obj_str = long_num_format(len(data))
        with TimedLog(self._logger.info, f"processed {n_obj_str} records"):
            limits = LimitTracker()
            patches: dict[int, PatchCatalog] = {}
            patch_iter = data.groupby(patch_name)
            if progress:
                patch_iter = job_progress_bar(patch_iter, total=n_patches)
            for patch_id, patch_data in patch_iter:
                if patch_id < 0:
                    raise ValueError("negative patch IDs are not supported")
                # drop extra columns
                patch_data = patch_data.drop(
                    columns=[col for col in patch_data.columns if col not in renames]
                )
                patch_data.rename(columns=renames, inplace=True)
                patch_data.reset_index(drop=True, inplace=True)
                # look up the center of the patch if given
                kwargs = dict(center=centers.get(patch_id))
                if unload:
                    # data will be written as feather file and loaded on demand
                    kwargs["cachefile"] = os.path.join(
                        cache_directory, f"patch_{patch_id:.0f}.feather"
                    )
                patch = PatchCatalog(int(patch_id), patch_data, **kwargs)
                limits.update(patch.redshifts)
                if unload:
                    patch.unload()
                patches[patch.id] = patch
            if progress:  # clean up if any patch was empty and skipped
                patch_iter.close()
            self._zmin, self._zmax = limits.get()
            self._patches = patches

        # also store the patch properties
        if unload:
            centers = self.centers.to_3d()
            property_df = pd.DataFrame(
                dict(
                    ids=self.ids,
                    x=centers.x,
                    y=centers.y,
                    z=centers.z,
                    r=self.radii.values,
                )
            )
            fpath = os.path.join(cache_directory, "properties.feather")
            property_df.to_feather(fpath)

    @classmethod
    def from_cache(cls, cache_directory: str, progress: bool = False) -> ScipyCatalog:
        super().from_cache(cache_directory)
        new = cls.__new__(cls)
        # load the patch properties
        fpath = os.path.join(cache_directory, "properties.feather")
        property_df = pd.read_feather(fpath)
        # transform data frame to dictionaries
        ids = property_df["ids"]
        centers = Coord3D.from_array(property_df[["x", "y", "z"]].to_numpy())
        radii = DistSky(property_df["r"].to_numpy())
        # transform to dictionary
        centers = {pid: center for pid, center in zip(ids, centers)}
        radii = {pid: radius for pid, radius in zip(ids, radii)}
        # load the patches
        limits = LimitTracker()
        new._patches = {}
        patch_files = list(os.listdir(cache_directory))
        if progress:
            patch_files = job_progress_bar(patch_files)
        for path in patch_files:
            if not path.startswith("patch"):
                continue
            abspath = os.path.join(cache_directory, path)
            if not os.path.isfile(abspath):
                continue
            patch_id = patch_id_from_path(path)
            patch = PatchCatalog.from_cached(
                abspath, center=centers.get(patch_id), radius=radii.get(patch_id)
            )
            limits.update(patch.redshifts)
            patch.unload()
            new._patches[patch.id] = patch
        new._zmin, new._zmax = limits.get()
        return new

    def __len__(self) -> int:
        return sum(len(patch) for patch in self._patches.values())

    def __getitem__(self, item: int) -> PatchCatalog:
        return self._patches[item]

    @property
    def ids(self) -> list[int]:
        return sorted(self._patches.keys())

    @property
    def n_patches(self) -> int:
        # seems ok to drop the last patch if that is empty and therefore missing
        return max(self._patches.keys()) + 1

    def __iter__(self) -> Iterator[PatchCatalog]:
        for patch_id in self.ids:
            patch = self._patches[patch_id]
            loaded = patch.is_loaded()
            patch.load()
            yield patch
            if not loaded:
                patch.unload()

    def is_loaded(self) -> bool:
        return all([patch.is_loaded() for patch in self._patches.values()])

    def load(self) -> None:
        super().load()
        for patch in self._patches.values():
            patch.load()

    def unload(self) -> None:
        super().unload()
        for patch in self._patches.values():
            patch.unload()

    def has_redshifts(self) -> bool:
        return all(patch.has_redshifts() for patch in self._patches.values())

    def has_weights(self) -> bool:
        return all(patch.has_weights() for patch in self._patches.values())

    @property
    def ra(self) -> NDArray[np.float_]:
        return np.concatenate([patch.ra for patch in iter(self)])

    @property
    def dec(self) -> NDArray[np.float_]:
        return np.concatenate([patch.dec for patch in iter(self)])

    @property
    def redshifts(self) -> NDArray[np.float_] | None:
        if self.has_redshifts():
            return np.concatenate([patch.redshifts for patch in iter(self)])
        else:
            return None

    @property
    def weights(self) -> NDArray[np.float_]:
        weights = []
        for patch in iter(self):
            if patch.has_weights():
                weights.append(patch.weights)
            else:
                weights.append(np.ones(len(patch)))
        return np.concatenate(weights)

    @property
    def patch(self) -> NDArray[np.int_]:
        return np.concatenate([np.full(len(patch), patch.id) for patch in iter(self)])

    def get_min_redshift(self) -> float:
        return self._zmin

    def get_max_redshift(self) -> float:
        return self._zmax

    @property
    def total(self) -> float:
        return self.get_totals().sum()

    def get_totals(self) -> NDArray[np.float_]:
        return np.array([patch.total for patch in self._patches.values()])

    @property
    def centers(self) -> CoordSky:
        return CoordSky.from_coords([self._patches[pid].center for pid in self.ids])

    @property
    def radii(self) -> DistSky:
        return DistSky.from_dists([self._patches[pid].radius for pid in self.ids])

    def correlate(
        self,
        config: Configuration,
        binned: bool,
        other: ScipyCatalog | None = None,
        linkage: PatchLinkage | None = None,
        progress: bool = False,
    ) -> NormalisedCounts | dict[str, NormalisedCounts]:
        super().correlate(config, binned, other, linkage)

        auto = other is None
        if not auto and not isinstance(other, ScipyCatalog):
            raise TypeError

        if linkage is None:
            cat_for_linkage = self
            if not auto:
                if len(other) > len(self):
                    cat_for_linkage = other
            linkage = PatchLinkage.from_setup(config, cat_for_linkage)
        patch1_list, patch2_list = linkage.get_patches(
            self, other, config.backend.crosspatch
        )
        n_jobs = len(patch1_list)

        # prepare job scheduling
        pool = ParallelHelper(
            function=_count_pairs_thread,
            n_items=n_jobs,
            num_threads=config.backend.get_threads(n_jobs),
        )
        # patch1: PatchCatalog
        pool.add_iterable(patch1_list)
        # patch2: PatchCatalog
        pool.add_iterable(patch2_list)
        # scales: Sequence[Scale]
        pool.add_constant(list(config.scales))
        # cosmology: TypeCosmology
        pool.add_constant(config.cosmology)
        # z_bins: NDArray[np.float_]
        pool.add_constant(config.binning.zbins)
        # bin1: bool
        pool.add_constant(self.has_redshifts())
        # bin2: bool
        pool.add_constant(binned if other is not None else True)
        # dist_weight_scale: float | None
        pool.add_constant(config.scales.rweight)
        # weight_res: int
        pool.add_constant(config.scales.rbin_num)

        binning = pd.IntervalIndex.from_breaks(config.binning.zbins)
        n_bins = len(binning)
        n_patches = self.n_patches
        # set up data to repack task results from [ids->scale] to [scale->ids]
        totals1 = np.zeros((n_patches, n_bins))
        totals2 = np.zeros((n_patches, n_bins))
        count_dict = {
            str(scale): PatchedCount.zeros(binning, n_patches, auto=auto)
            for scale in config.scales
        }
        # run the scheduled tasks
        result_iter = pool.iter_result(ordered=False)
        # add an optional progress bar
        if progress:
            result_iter = job_progress_bar(result_iter, total=pool.n_jobs())
        for patch_data in result_iter:
            id1, id2 = patch_data.patches
            # record total weight per bin, overwriting OK since identical
            totals1[id1] = patch_data.totals1
            totals2[id2] = patch_data.totals2
            # record counts at each scale
            for scale_key, count in patch_data.counts.items():
                if auto and id1 == id2:
                    count = count * 0.5  # autocorr. pairs are counted twice
                count_dict[scale_key].set_measurement((id1, id2), count)

        total = PatchedTotal(  # not scale-dependent
            binning=binning, totals1=totals1, totals2=totals2, auto=auto
        )
        return pack_results(count_dict, total)

    def true_redshifts(
        self,
        config: Configuration,
        sampling_config: ResamplingConfig | None = None,
        progress: bool = False,
    ) -> HistData:
        super().true_redshifts(config)
        if sampling_config is None:
            sampling_config = ResamplingConfig()  # default values

        if not self.has_redshifts():
            raise ValueError("catalog has no redshifts")
        # compute the reshift histogram in each patch
        pool = ParallelHelper(
            function=_histogram_thread,
            n_items=self.n_patches,
            num_threads=config.backend.get_threads(self.n_patches),
        )
        # patch: PatchCatalog
        pool.add_iterable(self._patches.values())
        # NDArray[np.float_]
        pool.add_constant(config.binning.zbins)
        iterator = pool.iter_result()
        if progress:
            iterator = job_progress_bar(iterator)
        hist_counts = np.array(list(iterator))

        # construct the output data samples
        binning = pd.IntervalIndex.from_breaks(config.binning.zbins)
        patch_idx = sampling_config.get_samples(self.n_patches)
        nz_data = hist_counts.sum(axis=0)
        nz_samp = np.sum(hist_counts[patch_idx], axis=1)
        return HistData(
            binning=binning,
            data=nz_data,
            samples=nz_samp,
            method=sampling_config.method,
        )



from __future__ import annotations

from deprecated import deprecated

from yaw.correlation import CorrData, CorrFunc
from yaw.correlation.paircounts import NormalisedCounts
from yaw.redshifts import HistData

__all__ = ["CorrelationData", "CorrelationFunction", "HistogramData", "PairCountResult"]


@deprecated(reason="renamed to yaw.CorrData", action="module", version="2.3.2")
class CorrelationData(CorrData):
    """
    .. deprecated:: 2.3.2
        Renamed to :meth:`yaw.CorrData`.
    """

    pass


@deprecated(reason="renamed to yaw.CorrFunc", action="module", version="2.3.2")
class CorrelationFunction(CorrFunc):
    """
    .. deprecated:: 2.3.2
        Renamed to :meth:`yaw.CorrFunc`.
    """

    pass


@deprecated(reason="renamed to yaw.HistData", action="module", version="2.3.2")
class HistogramData(HistData):
    """
    .. deprecated:: 2.3.2
        Renamed to :meth:`yaw.HistData`.
    """

    pass


@deprecated(
    reason="renamed to yaw.correlation.paircounts.NormalisedCounts",
    action="module",
    version="2.3.2",
)
class PairCountResult(NormalisedCounts):
    """
    .. deprecated:: 2.3.2
        Renamed to :obj:`yaw.correlation.paircounts.NormalisedCounts`.
    """

    pass



"""This module implements the computation of correlation functions with
:func:`crosscorrelate` and :func:`autocorrelate` from input data catalogues, see
:mod:`yaw.catalogs`.

These measurement methods return the main class, the correlation function
:obj:`CorrFunc` container. It holds the normalised pair counts (in
:obj:`~paircounts.NormalisedCounts` containers) computed within each spatial
patch and bins of redshift. The actual correlation function values and its
uncertainty (from resampling the spatial patches) can be computed using the
:meth:`CorrFunc.sample()`, which returns a :obj:`CorrData` container.

For the conversion of the correlation functions to a redshift estimate refer to
the :mod:`yaw.redshifts` module.
"""

from yaw.correlation.corrfuncs import (
    CorrData,
    CorrFunc,
    add_corrfuncs,
    autocorrelate,
    crosscorrelate,
)

__all__ = ["CorrData", "CorrFunc", "add_corrfuncs", "autocorrelate", "crosscorrelate"]



"""This module implements some math and array related functions.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, TypeVar

import numpy as np
from numpy.typing import NDArray

from ._math import _rebin

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import ArrayLike

    from yaw.core.containers import SampledData

__all__ = [
    "array_equal",
    "outer_triu_sum",
    "apply_bool_mask_ndim",
    "apply_slice_ndim",
    "sgn",
    "cov_from_samples",
    "global_covariance",
    "corr_from_cov",
    "rebin",
    "shift_histogram",
]


_Tarr = TypeVar("_Tarr", bound=NDArray)


def array_equal(arr1: NDArray, arr2: NDArray) -> bool:
    """Check if the shape and array elements of two numpy array are equal."""
    return (
        isinstance(arr1, np.ndarray)
        and isinstance(arr2, np.ndarray)
        and arr1.shape == arr2.shape
        and (arr1 == arr2).all()
    )


def outer_triu_sum(
    a: ArrayLike, b: ArrayLike, *, k: int = 0, axis: int | None = None
) -> NDArray:
    """Compute the sum over the upper triangle of the outer product.

    Shapes of input array must be identical. Equivalent to

    >>> np.triu(np.outer(a, b), k).sum(axis)

    but supports extra dimensions in a and b and does not construct the full
    outer product matrix in memory.

    Args:
        a (:obj:`NDArray`):
            First input array.
        b (:obj:`NDArray`):
            Second input array.

    Keyword args:
        k (:obj:`int`, optional):
            Diagonal above which to zero elements. `k = 0` (the default) is
            the main diagonal, `k < 0` is below it and `k > 0` is above.
        axis (:obj:`int`, optional):
            Array axis over which the outer product is summed. All by default.
    """
    a = np.atleast_1d(a)
    b = np.atleast_1d(b)
    if a.shape != b.shape:
        raise IndexError("shape of 'a' and 'b' does not match")
    # allocate output array
    dtype = (a[0] * b[0]).dtype  # correct dtype for product
    N = len(a)
    # sum all elements
    if axis is None:
        result = np.zeros_like(a[0], dtype=dtype)
        for i in range(min(N, N - k)):
            result += (a[i] * b[max(0, i + k) :]).sum(axis=0)
    # sum row-wise
    elif axis == 1:
        result = np.zeros_like(b, dtype=dtype)
        for i in range(min(N, N - k)):
            result[i] = (a[i] * b[max(0, i + k) :]).sum(axis=0)
    # sum column-wise
    elif axis == 0:
        result = np.zeros_like(a, dtype=dtype)
        for i in range(max(0, k), N):
            result[i] = (a[: min(N, max(0, i - k + 1))] * b[i]).sum(axis=0)
    return result[()]


def apply_bool_mask_ndim(
    array: _Tarr, mask: NDArray[np.bool_], axis: int | Sequence[int] | None = None
) -> _Tarr:
    """Apply a boolean mask (``mask``) to one or many axes of a numpy array."""
    if axis is None:
        axis = list(range(array.ndim))
    elif isinstance(axis, int):
        axis = [axis]
    result = array
    for ax in axis:
        if result.shape[ax] != len(mask):
            raise IndexError(
                f"boolean index did not match indexed array along dimension "
                f"{ax}; dimension is {result.shape[ax]} but corresponding "
                f"boolean dimension is {len(mask)}"
            )
        result = np.compress(mask, result, axis=ax)
    return result


def apply_slice_ndim(
    array: _Tarr, item: int | slice | Sequence, axis: int | Sequence[int] | None = None
) -> _Tarr:
    """Apply an integer subset or slice (``item``) to one or many axes of a
    numpy array."""
    if axis is None:
        axis = list(range(array.ndim))
    elif isinstance(axis, int):
        axis = [axis]
    if isinstance(item, slice):
        slices = [slice(None) for _ in range(array.ndim)]
        for ax in axis:
            slices[ax] = item
        return array[tuple(slices)]
    else:
        if isinstance(item, int):
            item = [item]
        indices = [range(n) for n in array.shape]
        for ax in axis:
            indices[ax] = item
        mesh_indices = np.ix_(*indices)
        return array[mesh_indices]


def sgn(val: ArrayLike) -> ArrayLike:
    """Compute the sign of a (array of) numbers, with positive numbers and 0
    returning 1, negative number returning -1."""
    return np.where(val == 0, 1.0, np.sign(val))


def cov_from_samples(
    samples: NDArray | Sequence[NDArray],
    method: str,
    rowvar: bool = False,
    kind: str = "full",  # full, diag, var
) -> NDArray:
    """Compute a joint covariance from a sequence of data samples.

    These samples can be jackknife or bootstrap samples (etc.). If more than one
    set of samples is provided, the samples are concatenated along the second
    axis (default) or along the first axis if ``rowvar=True``.

    Args:
        samples (:obj`:NDArray`, :obj:`Sequence[NDArray]`):
            One or many sets of data samples. The number of samples must be
            identical.
        method (:obj:`str`, optional):
            The resampling method that generated the samples, see
            :obj:`~yaw.config.options.Options.method`.
        rowvar (:obj:`bool`, optional):
            Whether the each row represents an observable. Determines the
            concatenation for multiple input sample sets.
        kind (:obj:`str`, optional):
            Determines the kind of covariance computed, see
            :obj:`~yaw.config.options.Options.kind`.
    """
    ax_samples = 1 if rowvar else 0
    ax_observ = 0 if rowvar else 1
    # if many samples are provided, concatenate them
    try:
        concat_samples = np.concatenate(samples, axis=ax_observ)
    except np.AxisError:
        concat_samples = samples

    # np.cov will produce a scalar value instead of matrix with shape (N,N)
    # for a single sample with shape (1,N)
    if concat_samples.shape[ax_samples] == 1:
        n_obs = concat_samples.shape[ax_observ]
        return np.full((n_obs, n_obs), np.nan)

    if method == "bootstrap":
        covmat = np.cov(concat_samples, rowvar=rowvar, ddof=1)
    elif method == "jackknife":
        n_samples = concat_samples.shape[ax_samples]
        covmat = np.cov(concat_samples, rowvar=rowvar, ddof=0) * (n_samples - 1)
    else:
        raise ValueError(f"invalid sampling method '{method}'")

    if kind == "full":
        pass
    elif kind == "diag":
        # get a matrix with only the main diagonal elements
        idx_diag = 0
        cov_diags = np.diag(np.diag(covmat, k=idx_diag), k=idx_diag)
        try:
            for sample in samples:
                # go to next diagonal that contains correlations between samples
                idx_diag += sample.shape[ax_observ]
                # add just the diagonal values to the existing matrix
                cov_diags += np.diag(np.diag(covmat, k=-idx_diag), k=-idx_diag)
                cov_diags += np.diag(np.diag(covmat, k=idx_diag), k=idx_diag)
        except IndexError:
            raise
        covmat = cov_diags
    elif kind == "var":
        covmat = np.diag(np.diag(covmat, k=0), k=0)
    else:
        raise ValueError(f"invalid covariance kind '{kind}'")
    return covmat


def global_covariance(
    data: Sequence[SampledData], method: str | None = None, kind: str = "full"
) -> NDArray:
    """Compute a joint covariance from a set of resampled data.

    Typically applied to a set of :obj:`CorrData`,
    :obj:`~yaw.redshifts.RedshiftData`, or :obj:`~yaw.redshifts.HistData`
    containers. The joint covariance is computed by concatenating the samples
    along the redshift binning axis.

    .. Warning::
        The input containers must have the same number of samples, and use the
        same resampling method. They also should be of the same type.

    Args:
        data (sequence of :obj:`SampledData`):
            The input containers, should be of the same type.
        method (:obj:`str`, optional):
            Specify the sampling method to use. All other containers must follow
            this convention.
        kind (:obj:`str`, optional):
            The method to compute the covariance matrix, see
            :func:`yaw.core.math.cov_from_samples`.

    Returns:
        :obj:`NDArray`:
            Jointly estimated covariance. Dimension matches the sum of all
            redshift bins in the input containers.
    """
    if len(data) == 0:
        raise ValueError("'data' must be a sequence with at least one item")
    if method is None:
        method = data[0].method
    for d in data[1:]:
        if d.method != method:
            raise ValueError("resampling method of data items is inconsistent")
    return cov_from_samples([d.samples for d in data], method=method, kind=kind)


def corr_from_cov(covariance: NDArray) -> NDArray:
    """Convert an input covariance matrix to a covariance matrix."""
    v = np.sqrt(np.diag(covariance))
    outer_v = np.outer(v, v)
    return covariance / outer_v


def rebin(
    bins_new: NDArray[np.float_],
    bins_old: NDArray[np.float_],
    counts_old: NDArray[np.float_],
) -> NDArray[np.float_]:
    """Recompute compute histogram counts for a new binning.

    The new counts are computed by summing the fractional contribution of the
    counts from the original binning to the binning. The new binning may exceed
    or just partially cover the range of the original binning.

    Args:
        bins_new (:obj:`NDArray`):
            The new bin edges on which the counts are reevaluated.
        bins_old (:obj:`NDArray`):
            The bin edges from which the original counts were computed.
        counts_old (:obj:`NDArray`):
            The original histogram counts.

    Returns:
        :obj:`NDArray`:
            The histogram counts for the new binning.

    .. Note::
        Implemented as C extension.
    """
    return _rebin(
        bins_new.astype(np.float_),
        bins_old.astype(np.float_),
        counts_old.astype(np.float_),
    )


def shift_histogram(
    bins: NDArray, counts: NDArray, *, A: float = 1.0, dx: float = 0.0
) -> NDArray:
    """Shift a histogram by a fixed value.

    The histogram values are recomputed for new bin edges that are shifted by
    the desired amount using :func:`rebin`. The normalisation of the histogram
    may change if the shifted binning does not cover the full range of the data.

    Args:
        bins (:obj:`NDArray`):
            The bin edges from which the counts were computed.
        counts_old (:obj:`NDArray`):
            The histogram counts.

    Keyword args:
        A (:obj:`float`, optional):
            Scalar amplitude used to rescale the new histgram counts.
        dx (:obj:`float`, optional):
            Amount by which the histogram (i.e. the bin edges) are shifted.

    Returns:
        :obj:`NDArray`:
            The shifted histogram counts.
    """
    bins_old = bins.astype(np.float_)
    bins_new = bins_old + dx
    return A * _rebin(bins_new, bins_old, counts.astype(np.float_))



from __future__ import annotations

from typing import TYPE_CHECKING

from yaw.catalogs.catalog import BackendError, BaseCatalog

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame

    from yaw.core.coordinates import Coordinate

__all__ = ["NewCatalog"]


class NewCatalog:
    """Factory class for data catalogues implemented by the backends.

    A catalogue provides all the functionality to compute pair counts for
    correlation measurements by implementing an interface to the object
    positions, spatial patches for error estimation, and data management if the
    data is cached on disk. Aside from accessing the data directly, the most
    important methods are the :meth:`correlate` (pair counting) and
    :meth:`true_redshifts` (redshift histogram, if redshifts are provided).

    A new catalogue can be created using an instance of this factory class.
    The sole argument is the name of the backend for which catalogue instances
    should be produced. For example

    >>> yaw.NewCatalog("scipy")
    NewCatalog<scipy>()

    is the default factory, which produces catalogues for the ``scipy`` backend
    through its constructor methods.

    A key concept is :ref:`caching<caching>`, which can be used to reduce
    memory usage or even speed up the computation for some backends. A cache
    directory is a directory in which temporary data is stored in different
    formats (depending on the backend), such that parts of the data (typically
    individual spatial patches) can be read back into memory on demand.

    .. Warning::

        - The ``scipy`` backend does not preserve the order the input data, but
          instead groups objects by there spatial patch.
        - The ``treecorr`` backend does currently not support restoration from
          cache.
    """

    def __init__(self, backend: str = "scipy") -> None:
        """Create a new catalogue factory.

        Args:
            backend (:obj:`str`):
                Specify the backend for which the catalog instances should be
                produced for. For availble options see
                :attr:`~yaw.config.options.Options.backend`.
        """
        try:
            self.catalog: BaseCatalog = BaseCatalog._backends[backend]
            self.backend_name = backend
        except KeyError as e:
            raise BackendError(f"invalid backend '{backend}'") from e

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.backend_name}>()"

    def from_dataframe(
        self,
        data: DataFrame,
        ra_name: str,
        dec_name: str,
        *,
        patch_name: str | None = None,
        patch_centers: BaseCatalog | Coordinate | None = None,
        n_patches: int | None = None,
        redshift_name: str | None = None,
        weight_name: str | None = None,
        cache_directory: str | None = None,
        progress: bool = False,
    ) -> BaseCatalog:
        """Build a catalogue from in-memory data.

        Specify the names of the required and or available columns in a
        :obj:`pandas.DataFrame`. Additional parameters control the creation
        spatial patches used for error estimates. Patches can be assigned based
        on a column in the data frame (``patch_name``), constructed from a set
        of existing patch centers (``patch_centers``), or generated with
        `k`-means clustering (``n_patches``).

        Args:
            data (:obj:`pandas.Dataframe`):
                Holds the catalog data.
            ra_name (:obj:`str`):
                Name of the column with right ascension data in degrees.
            dec_name (:obj:`str`):
                Name of the column with declination data in degress.

        Keyword Args:
            patch_name (:obj:`str`, optional):
                Name of the column that specifies the patch index, i.e.
                assigning each object to a spatial patch. Index starts counting
                from 0 (see :ref:`patches`).
            patch_centers (:obj:`~yaw.catalogs.BaseCatalog`, :obj:`~yaw.core.coordinates.Coordinate`, optional):
                Assign objects to existing patch centers based on their
                coordinates. Must be either a different catalog instance or a
                vector of coordinates.
            n_patches (:obj:`int`, optional):
                Assign objects to a given number of patches, generated using
                k-means clustering.
            redshift_name (:obj:`str`, optional):
                Name of the column with point-redshift estimates.
            weight_name (:obj:`str`, optional):
                Name of the column with object weights.
            cache_directory (:obj:`str`, optional):
                Path to directory used to cache patch data, must exists (see
                :ref:`caching`). If provided, patch data is automatically
                unloaded from memory.
            progress (:obj:`bool`, optional):
                Display a progress bar while creating patches.

        .. Note::
            Either of ``patch_name``, ``patch_centers``, or ``n_patches`` is
            required.

        Caching may significantly speed up parallel computations (e.g.
        :meth:`correlate`), accessing data attributes will trigger loading
        cached data as long as the catalog remains in the unloaded state (see
        :meth:`load` and :meth:`unload`).

        The underlying patch data can be accessed through indexing and
        iterating the Catalog instance.

        .. Note::
            TODO: Provide an example.
        """
        return self.catalog(
            data,
            ra_name,
            dec_name,
            patch_name=patch_name,
            patch_centers=patch_centers,
            n_patches=n_patches,
            redshift_name=redshift_name,
            weight_name=weight_name,
            cache_directory=cache_directory,
            progress=progress,
        )

    def from_file(
        self,
        filepath: str,
        patches: str | int | BaseCatalog | Coordinate,
        ra: str,
        dec: str,
        *,
        redshift: str | None = None,
        weight: str | None = None,
        sparse: int | None = None,
        cache_directory: str | None = None,
        file_ext: str | None = None,
        progress: bool = False,
        **kwargs,
    ) -> BaseCatalog:
        """
        Build catalogue from data file.

        Loads the input file and constructs the catalogue using the specified
        column names.

        Args:
            filepath (:obj:`str`):
                Path to the input data file.
            patches (:obj:`str`, :obj:`int`, :obj:`~yaw.catalogs.BaseCatalog`, :obj:`~yaw.core.coordinates.Coordinate`):
                Specifies the construction of patches. If `str`, patch indices
                are read from the file. If `int`, generates this number of
                patches. Otherwise assign objects based on existing patch
                centers from a catalog instance or a coordinate vector.
            ra (:obj:`str`):
                Name of the column with right ascension data in degrees.
            dec (:obj:`str`):
                Name of the column with declination data in degress.

        Keyword Args:
            redshift (:obj:`str`, optional):
                Name of the column with point-redshift estimates.
            weight (:obj:`str`, optional):
                Name of the column with object weights.
            sparse (:obj:`int`, optional):
                Load every N-th row of the input data.
            cache_directory (:obj:`str`, optional):
                Path to directory used to cache patch data, must exists (see
                :ref:`caching`). If provided, patch data is automatically
                unloaded from memory.
            file_ext (:obj:`str`, optional):
                Hint for the input file type, if a uncommon file extension is
                used.
            progress (:obj:`bool`, optional):
                Display a progress bar while creating patches.

        Returns:
            :obj:`BaseCatalog`

        .. Note::
            Currently, the following file extensions are recognised
            automatically:

            - FITS: ``.fits``, ``.cat``
            - CSV: ``.csv``
            - HDF5: ``.hdf5``, ``.h5``,
            - Parquet: ``.pqt``, ``.parquet``
            - Feather: ``.feather``

            Otherwise provide the appropriate extension (including the dot)
            in the ``file_ext`` argument.
        """
        return self.catalog.from_file(
            filepath,
            patches,
            ra,
            dec,
            redshift=redshift,
            weight=weight,
            sparse=sparse,
            cache_directory=cache_directory,
            file_ext=file_ext,
            progress=progress,
            **kwargs,
        )

    def from_cache(self, cache_directory: str, progress: bool = False) -> BaseCatalog:
        """
        Restore the catalogue from its cache directory.

        Args:
            cache_directory (:obj:`str`):
                Path to the cache directory.
            progress (:obj:`bool`, optional):
                Display a progress bar while restoring patches.

        Returns:
            :obj:`BaseCatalog`
        """
        return self.catalog.from_cache(cache_directory, progress=progress)



from __future__ import annotations

from typing import NoReturn, get_args

import astropy.cosmology

from yaw.core.cosmology import TypeCosmology, get_default_cosmology

__all__ = [
    "cosmology_to_yaml",
    "yaml_to_cosmology",
    "parse_cosmology",
    "parse_section_error",
]


class ConfigError(Exception):
    pass


def cosmology_to_yaml(cosmology: TypeCosmology) -> str:
    """Try to represent the cosmological model in a YAML-friendly way.

    If it is one of the names :obj:`astropy` cosmologies, returns the name,
    otherwise raises an :exc:`ConfigError`."""
    if not isinstance(cosmology, astropy.cosmology.FLRW):
        raise ConfigError("cannot serialise custom cosmoligies to YAML")
    if cosmology.name not in astropy.cosmology.available:
        raise ConfigError("can only serialise predefined astropy cosmologies to YAML")
    return cosmology.name


def yaml_to_cosmology(cosmo_name: str) -> TypeCosmology:
    """Reinstantiate the cosmological model from its name representation."""
    if cosmo_name not in astropy.cosmology.available:
        raise ConfigError(
            f"unknown cosmology with name '{cosmo_name}', see "
            "'astropy.cosmology.available'"
        )
    return getattr(astropy.cosmology, cosmo_name)


def parse_cosmology(cosmology: TypeCosmology | str | None) -> TypeCosmology:
    """Construct the cosmological model.

    Either returns the default model, loads one of the named cosmological
    :obj:`astropy` cosmologies, otherwise returns the input if it is an
    :obj:`astropy` cosmologies or subclass of
    :obj:`yaw.core.cosmology.CustomCosmology`."""
    if cosmology is None:
        cosmology = get_default_cosmology()
    elif isinstance(cosmology, str):
        cosmology = yaml_to_cosmology(cosmology)
    elif not isinstance(cosmology, get_args(TypeCosmology)):
        which = ", ".join(get_args(TypeCosmology))
        raise ConfigError(f"'cosmology' must be instance of: {which}")
    return cosmology


def parse_section_error(
    exception: Exception, section: str, reraise: Exception = ConfigError
) -> NoReturn:
    """Reraises are a more descriptive exception from an existing exception when
    parsing a YAML configuration.

    Covered cases are undefined key names, missing required key names or
    entirely missing subsection in the configuration.
    """
    msg = exception.args[0]
    item = msg.split("'")[1]
    if isinstance(exception, TypeError):
        if "__init__() got an unexpected keyword argument" in msg:
            raise reraise(
                f"encountered unknown option '{item}' in section '{section}'"
            ) from exception
        elif "missing" in msg:
            raise reraise(
                f"missing option '{item}' in section '{section}'"
            ) from exception
    elif isinstance(exception, KeyError):
        raise reraise(f"missing section '{section}'") from exception
    raise



from __future__ import annotations

import logging
from dataclasses import asdict, dataclass, field
from typing import TYPE_CHECKING, Any, get_args

import numpy as np
import yaml

from yaw.config import OPTIONS
from yaw.config import default as DEFAULT
from yaw.config import utils
from yaw.config.backend import BackendConfig
from yaw.config.binning import (
    AutoBinningConfig,
    ManualBinningConfig,
    make_binning_config,
    warn_binning_args_ignored,
)
from yaw.config.scales import ScalesConfig
from yaw.core.abc import DictRepresentation
from yaw.core.cosmology import TypeCosmology, get_default_cosmology, r_kpc_to_angle
from yaw.core.docs import Parameter

if TYPE_CHECKING:  # pragma: no cover
    from matplotlib.figure import Figure
    from numpy.typing import ArrayLike, NDArray

    from yaw.catalogs import BaseCatalog
    from yaw.core.utils import TypePathStr

__all__ = ["Configuration"]


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Configuration(DictRepresentation):
    """The central configration for correlation measurements.

    Bundles the configuration of measurement scales, redshift binning, and
    backend parameters in a single, hierarchical configuration class.
    Additionally holds the cosmological model used for distance calculations.

    .. Note::

        The structure and meaning of the parameters is described in more detail
        in the specialised configuration objects :obj:`ScalesConfig`,
        :obj:`AutoBinningConfig` / :obj:`ManualBinningConfig`,
        :obj:`BackendConfig`, which are stored as class attributes
        :obj:`scales`, :obj:`binning`, and :obj:`backend`.

        To access e.g. the lower measurement scale limit, use

        >>> Configuration.scales.rmin
        ...

        which accesses the :obj:`ScalesConfig.rmin` attribute.

    A new instance should be constructed with the :meth:`create` method or
    as a modified variant with the :meth:`modify` method.

    Args:
        scales (:obj:`~yaw.config.ScalesConfig`):
            The configuration of the measurement scales.
        binning (:obj:`~yaw.config.AutoBinningConfig`, :obj:`~yaw.config.ManualBinningConfig`):
            The redshift binning configuration.
        backend (:obj:`~yaw.config.BackendConfig`):
            The backend-specific configuration.
        cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`~yaw.core.cosmology.CustomCosmology`, :obj:`str`, :obj:`None`, optional)
            The cosmological model for distance calculations.
    """

    scales: ScalesConfig
    """The configuration of the measurement scales."""
    binning: AutoBinningConfig | ManualBinningConfig
    """The redshift binning configuration."""
    backend: BackendConfig = field(default_factory=BackendConfig)
    """The backend-specific configuration."""
    cosmology: TypeCosmology | str | None = field(
        default=DEFAULT.Configuration.cosmology,
        metadata=Parameter(
            type=str,
            choices=OPTIONS.cosmology,
            help="cosmological model used for distance calculations",
            default_text="(see astropy.cosmology, default: %(default)s)",
        ),
    )
    """The cosmological model for distance calculations."""

    def __post_init__(self) -> None:
        # parse cosmology
        if self.cosmology is None:
            cosmology = get_default_cosmology()
        elif isinstance(self.cosmology, str):
            cosmology = utils.yaml_to_cosmology(self.cosmology)
        elif not isinstance(self.cosmology, get_args(TypeCosmology)):
            which = ", ".join(get_args(TypeCosmology))
            raise utils.ConfigError(f"'cosmology' must be instance of: {which}")
        else:
            cosmology = self.cosmology
        cosmology = utils.parse_cosmology(self.cosmology)
        object.__setattr__(self, "cosmology", cosmology)

    @classmethod
    def create(
        cls,
        *,
        cosmology: TypeCosmology | str | None = DEFAULT.Configuration.cosmology,
        # ScalesConfig
        rmin: ArrayLike,
        rmax: ArrayLike,
        rweight: float | None = DEFAULT.Configuration.scales.rweight,
        rbin_num: int = DEFAULT.Configuration.scales.rbin_num,
        # AutoBinningConfig / ManualBinningConfig
        zmin: ArrayLike = None,
        zmax: ArrayLike = None,
        zbin_num: int | None = DEFAULT.Configuration.binning.zbin_num,
        method: str = DEFAULT.Configuration.binning.method,
        zbins: NDArray[np.float_] | None = None,
        # BackendConfig
        thread_num: int | None = DEFAULT.Configuration.backend.thread_num,
        crosspatch: bool = DEFAULT.Configuration.backend.crosspatch,
        rbin_slop: float = DEFAULT.Configuration.backend.rbin_slop,
    ) -> Configuration:
        """Create a new configuration object.

        Except for the ``cosmology`` parameter, all other parameters are passed
        to the constructors of the respective :obj:`ScalesConfig`,
        :obj:`AutoBinningConfig` / :obj:`ManualBinningConfig`,
        :obj:`BackendConfig` classes.

        .. Note::

            If custom bin edges are provided through the ``zbins`` parameter,
            ``zmin``, ``zmax``, ``zbin_num`` (optional), and ``method``
            (optional) are ignored. Otherwise, at least ``zmin``, ``zmax`` are
            required and a binning will be generated automatically.

        Otherwise, only ``rmin`` and ``rmax`` are required arguments, e.g.:

        >>> yaw.Configuration.create(rmin=100, rmax=1000, zmin=0.1, zmax=1.0)

        Keyword Args:
            cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`~yaw.core.cosmology.CustomCosmology`, :obj:`str`, :obj:`None`, optional):
                Named astropy cosmology used to compute distances. For options
                see :obj:`~yaw.config.options.Options.cosmology`.
            rmin (:obj:`ArrayLike`):
                (List of) lower scale limit in kpc (pyhsical).
            rmax (:obj:`ArrayLike`):
                (List of) upper scale limit in kpc (pyhsical).
            rweight (:obj:`float`, optional):
                Weight galaxy pairs by their separation to power 'rweight'.
            rbin_num (:obj:`int`, optional):
                Number of bins in log r used (i.e. resolution) to compute
                distance weights.
            zmin (:obj:`float`):
                Lower redshift limit.
            zmax (:obj:`float`):
                Upper redshift limit.
            zbin_num (:obj:`int`, optional):
                Number of redshift bins
            method (:obj:`str`, optional):
                Method used to generate the redshift binning. For options see
                :obj:`~yaw.config.options.Options.binning`.
            zbins (:obj:`NDArray`, optional):
                Manually define redshift bin edges.
            thread_num (:obj:`int`, optional):
                Default number of threads to use.
            crosspatch (:obj:`bool`, optional):
                whether to count pairs across patch boundaries (scipy backend
                only)
            rbin_slop (:obj:`float`, optional):
                TreeCorr 'rbin_slop' parameter

        Returns:
            :obj:`Configuration`
        """
        cosmology = utils.parse_cosmology(cosmology)
        scales = ScalesConfig(rmin=rmin, rmax=rmax, rweight=rweight, rbin_num=rbin_num)
        binning = make_binning_config(
            cosmology=cosmology,
            zmin=zmin,
            zmax=zmax,
            zbin_num=zbin_num,
            method=method,
            zbins=zbins,
        )
        backend = BackendConfig(
            thread_num=thread_num, crosspatch=crosspatch, rbin_slop=rbin_slop
        )
        return cls(scales=scales, binning=binning, backend=backend, cosmology=cosmology)

    def modify(
        self,
        *,
        cosmology: TypeCosmology | str | None = DEFAULT.NotSet,
        # ScalesConfig
        rmin: ArrayLike | None = DEFAULT.NotSet,
        rmax: ArrayLike | None = DEFAULT.NotSet,
        rweight: float | None = DEFAULT.NotSet,
        rbin_num: int | None = DEFAULT.NotSet,
        # AutoBinningConfig /  ManualBinningConfig
        zmin: float | None = DEFAULT.NotSet,
        zmax: float | None = DEFAULT.NotSet,
        zbin_num: int | None = DEFAULT.NotSet,
        method: str | None = DEFAULT.NotSet,
        zbins: NDArray[np.float_] | None = DEFAULT.NotSet,
        # BackendConfig
        thread_num: int | None = DEFAULT.NotSet,
        crosspatch: bool | None = DEFAULT.NotSet,
        rbin_slop: float | None = DEFAULT.NotSet,
    ) -> Configuration:
        """Create a copy of the current configuration with updated parameter
        values.

        The method arguments are identical to :meth:`create`. Values that should
        not be modified are by default represented by the special value
        :obj:`~yaw.config.default.NotSet`.
        """
        config = self.to_dict()
        if cosmology is not DEFAULT.NotSet:
            if isinstance(cosmology, str):
                cosmology = utils.yaml_to_cosmology(cosmology)
            config["cosmology"] = utils.cosmology_to_yaml(cosmology)
        # ScalesConfig
        if rmin is not DEFAULT.NotSet:
            config["scales"]["rmin"] = rmin
        if rmax is not DEFAULT.NotSet:
            config["scales"]["rmax"] = rmax
        if rweight is not DEFAULT.NotSet:
            config["scales"]["rweight"] = rweight
        if rbin_num is not DEFAULT.NotSet:
            config["scales"]["rbin_num"] = rbin_num
        # AutoBinningConfig /  ManualBinningConfig
        if zbins is not DEFAULT.NotSet:
            warn_binning_args_ignored(zmin, zmax, zbin_num)
            config["binning"]["zbins"] = zbins
        else:
            if zmin is not DEFAULT.NotSet:
                config["binning"]["zmin"] = zmin
            if zmax is not DEFAULT.NotSet:
                config["binning"]["zmax"] = zmax
            if zbin_num is not DEFAULT.NotSet:
                config["binning"]["zbin_num"] = zbin_num
            if method is not DEFAULT.NotSet:
                config["binning"]["method"] = method
        # BackendConfig
        if thread_num is not DEFAULT.NotSet:
            config["backend"]["thread_num"] = thread_num
        if crosspatch is not DEFAULT.NotSet:
            config["backend"]["crosspatch"] = crosspatch
        if rbin_slop is not DEFAULT.NotSet:
            config["backend"]["rbin_slop"] = rbin_slop
        return self.__class__.from_dict(config)

    def plot_scales(
        self, catalog: BaseCatalog, log: bool = True, legend: bool = True
    ) -> Figure:
        """Plot the configured correlation scales at different redshifts in
        comparison to the size of patches in a data catalogue."""
        import matplotlib.pyplot as plt

        fig, ax_scale = plt.subplots(1, 1)
        # plot scale of annulus
        for r_min, r_max in self.scales.as_array():
            ang_min, ang_max = np.transpose(
                [
                    r_kpc_to_angle([r_min, r_max], z, self.cosmology)
                    for z in self.binning.zbins
                ]
            )
            ax_scale.fill_between(
                self.binning.zbins,
                ang_min,
                ang_max,
                step="post",
                alpha=0.3,
                label=rf"${r_min:.0f} < r \leq {r_max:.0f}$ kpc",
            )
        if legend:
            ax_scale.legend(loc="lower right")
        # plot patch sizes
        ax_patch = ax_scale.twiny()
        bins = np.histogram_bin_edges(catalog.radii)
        if log:
            ax_patch.set_yscale("log")
            bins = np.logspace(
                np.log10(bins[0]), np.log10(bins[-1]), len(bins), base=10.0
            )
        ax_patch.hist(
            catalog.radii, bins, orientation="horizontal", color="k", alpha=0.5
        )
        # decorate
        ax_scale.set_xlim(self.binning.zmin, self.binning.zmax)
        ax_scale.set_ylabel("Radius / rad")
        ax_scale.set_xlabel("Redshift")
        ax_patch.set_xlabel("Patch count")
        return fig

    @classmethod
    def from_dict(cls, the_dict: dict[str, Any], **kwargs) -> Configuration:
        config = {k: v for k, v in the_dict.items()}
        cosmology = utils.parse_cosmology(
            config.pop("cosmology", DEFAULT.Configuration.cosmology)
        )
        # parse the required subgroups
        try:
            scales = ScalesConfig.from_dict(config.pop("scales"))
        except (TypeError, KeyError) as e:
            utils.parse_section_error(e, "scales")
        try:
            binning_conf = config.pop("binning")
            if "zbins" in binning_conf:
                binning = ManualBinningConfig(binning_conf["zbins"])
            else:
                binning = AutoBinningConfig.generate(
                    cosmology=cosmology, **binning_conf
                )
        except (TypeError, KeyError) as e:
            utils.parse_section_error(e, "binning")
        # parse the optional subgroups
        try:
            backend = BackendConfig.from_dict(config.pop("backend"))
        except KeyError:
            backend = BackendConfig()
        except TypeError as e:
            utils.parse_section_error(e, "backend")
        # check that there are no entries left
        if len(config) > 0:
            key = next(iter(config.keys()))
            raise utils.ConfigError(f"encountered unknown section '{key}'")
        return cls(scales=scales, binning=binning, backend=backend, cosmology=cosmology)

    def to_dict(self) -> dict[str, Any]:
        values = dict()
        for attr in asdict(self):
            value = getattr(self, attr)  # avoid asdict() recursion
            if attr == "cosmology":
                values[attr] = utils.cosmology_to_yaml(value)
            else:
                values[attr] = value.to_dict()
        return values

    @classmethod
    def from_yaml(cls, path: TypePathStr) -> Configuration:
        """Create a new instance by loading the configuration from a YAML file.

        Args:
            path (:obj:`pathlib.Path`, :obj:`str`):
                Path to the YAML file containing the configuration.

        Returns:
            :obj:`Configuration`
        """
        logger.info(f"reading configuration file '{path}'")
        with open(str(path)) as f:
            config = yaml.safe_load(f.read())
        return cls.from_dict(config)

    def to_yaml(self, path: TypePathStr) -> None:
        """Store the configuration as YAML file.

        Args:
            path (:obj:`pathlib.Path`, :obj:`str`):
                Path to which the YAML file is written.
        """
        logger.info(f"writing configuration file '{path}'")
        string = yaml.dump(self.to_dict())
        with open(str(path), "w") as f:
            f.write(string)



from __future__ import annotations

import gc
from collections.abc import Iterator
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy.cluster import vq

from yaw.catalogs.scipy.kdtree import SphericalKDTree
from yaw.core.coordinates import Coord3D, Coordinate, CoordSky, Distance, DistSky

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray
    from pandas import DataFrame, Interval

__all__ = ["PatchCatalog"]


class NotAPatchFileError(Exception):
    pass


class CachingError(Exception):
    pass


def patch_id_from_path(fpath: str) -> int:
    """Extract the patch ID from the file name in the cache directory"""
    ext = ".feather"
    if not fpath.endswith(ext):
        raise NotAPatchFileError("input must be a .feather file")
    prefix, patch_id = fpath[: -len(ext)].rsplit("_", 1)
    return int(patch_id)


class PatchCatalog:
    """Represents a single spatial patch of a :obj:`ScipyCatalog`.

    A patch holds the data from single patch of the catalogue and provides
    method to access this data. Furthermore, it implements the caching to and
    restoring from disk. Data is temporarily saved to a .feather file, making
    it easy to pass patches to new threads and processes, which can, load the
    data back into memory if necessary.
    """

    id = 0
    """Unique index of the patch."""
    cachefile = None
    """The patch to the cached .feather data file if caching is enabled."""
    _data = pd.DataFrame()
    _len = 0
    _total = None
    _has_z = False
    _has_weights = False
    _center = None
    _radius = None

    def __init__(
        self,
        id: int,
        data: DataFrame,
        cachefile: str | None = None,
        center: Coordinate | None = None,
        radius: Distance | None = None,
        degrees: bool = True,
    ) -> None:
        """Create a new patch from a data frame.

        Coordiantes are converted to radian. If a cache path is provided, a
        cache file is created and the data is dropped from memory.

        Args:
            id (:obj:`int`):
                Unique index of the patch.
            data (:obj:`pandas.DataFrame`):
                Data frame with columns ``ra``, ``dec`` (by default assumed to
                be in degrees) and optionally ``weights``, ``redshift`` if
                either data is available.
            cachefile (:obj:`str`, optional):
                If provided, the data is cached as .feather file at this path.
            center (:obj:`yaw.core.coordiante.Coordiante`, optional):
                Center coordinates of the patch. Computed automatically if not
                provided.
            radius (:obj:`yaw.core.coordiante.Distance`, optional):
                The angular size of the patch. Computed automatically if not
                provided.
            degrees (:obj:`bool`):
                Whether the input coordinates ``ra``, ``dec`` are in degrees.
        """
        self.id = id
        if "ra" not in data:
            raise KeyError("right ascension column ('ra') is required")
        if "dec" not in data:
            raise KeyError("declination column ('dec') is required")
        if not set(data.columns) <= set(["ra", "dec", "redshift", "weights"]):
            raise KeyError(
                "'data' contains unidentified columns, optional columns are "
                "restricted to 'redshift' and 'weights'"
            )
        # next line is crucial, otherwise lines below modify data inplace
        self._data = data.copy()
        if degrees:
            self._data["ra"] = np.deg2rad(data["ra"])
            self._data["dec"] = np.deg2rad(data["dec"])
        # if there is a file path, store the file
        if cachefile is not None:
            self.cachefile = cachefile
            self._data.to_feather(cachefile)
        self._init(center, radius)

    def _init(
        self, center: Coordinate | None = None, radius: Distance | None = None
    ) -> None:
        self._len = len(self._data)
        self._has_z = "redshift" in self._data
        self._has_weights = "weights" in self._data
        if self.has_weights():
            self._total = float(self.weights.sum())
        else:
            self._total = len(self)

        # precompute (estimate) the patch center and size since it is quite fast
        # and the data is still loaded
        if center is None or radius is None:
            SUBSET_SIZE = 1000  # seems a reasonable, fast but not too sparse
            if self._len < SUBSET_SIZE:
                positions = self.pos.to_3d()
            else:
                rng = np.random.default_rng(seed=12345)
                which = rng.integers(0, self._len, size=SUBSET_SIZE)
                positions = self.pos[which].to_3d()

        # store in xyz coordinates
        if center is None:
            self._center = positions.mean()
        else:
            self._center = center.to_3d()

        if center is None or radius is None:  # new center requires recomputing
            # compute maximum distance to any of the data points
            radius = positions.distance(self._center).max()

        # store radius in radians
        self._radius = radius.to_sky()

    def __repr__(self) -> str:
        s = self.__class__.__name__
        s += f"(id={self.id}, length={len(self)}, loaded={self.is_loaded()})"
        return s

    def __len__(self) -> int:
        return self._len

    @classmethod
    def from_cached(
        cls,
        cachefile: str,
        center: Coordinate | None = None,
        radius: Distance | None = None,
    ) -> PatchCatalog:
        """Restore the patch instance from its cache file.

        Optionally, the center and radius of the patch can be provided to avoid
        recomputing these quantities.

        Args:
            cachefile (:obj:`str`):
                Path to the cach file (.feather)
            center (:obj:`yaw.core.coordiante.Coordiante`, optional):
                Center coordinates of the patch. Computed automatically if not
                provided.
            radius (:obj:`yaw.core.coordiante.Distance`, optional):
                The angular size of the patch. Computed automatically if not
                provided.
        """
        # create the data instance
        new = cls.__new__(cls)
        new.id = patch_id_from_path(cachefile)
        new.cachefile = cachefile
        try:
            new._data = pd.read_feather(cachefile)
        except Exception as e:
            args = ()
            if hasattr(e, "args"):
                args = e.args
            raise NotAPatchFileError(*args) from e
        new._init(center, radius)
        return new

    def is_loaded(self) -> bool:
        """Whether the data is present in memory"""
        return self._data is not None

    def require_loaded(self) -> None:
        """Raise a :obj:`CachingError` if the data is not present in memory."""
        if not self.is_loaded():
            raise CachingError("data is not loaded")

    def load(self, use_threads: bool = True) -> None:
        """Load the data from the cache file into memory.

        Raises a :obj:`CachingError` if no cache file is sepcified."""
        if not self.is_loaded():
            if self.cachefile is None:
                raise CachingError("no datapath provided to load the data")
            self._data = pd.read_feather(self.cachefile, use_threads=use_threads)

    def unload(self) -> None:
        """Drop the data from memory.

        Raises a :obj:`CachingError` if no cache file is sepcified."""
        if self.cachefile is None:
            raise CachingError("no datapath provided to unload the data")
        self._data = None
        gc.collect()

    def has_redshifts(self) -> bool:
        """Whether the patch data include redshifts."""
        return self._has_z

    def has_weights(self) -> bool:
        """Whether the patch data include weights."""
        return self._has_weights

    @property
    def data(self) -> DataFrame:
        """Direct access to the underlying :obj:`pandas.DataFrame` which holds
        the patch data."""
        self.require_loaded()
        return self._data

    @property
    def ra(self) -> NDArray[np.float_]:
        """Get an array of the right ascension values in radians.

        Raises a :obj:`CachingError` if data is not loaded."""
        self.require_loaded()
        return self._data["ra"].to_numpy()

    @property
    def dec(self) -> NDArray[np.float_]:
        """Get an array of the declination values in radians.

        Raises a :obj:`CachingError` if data is not loaded."""
        self.require_loaded()
        return self._data["dec"].to_numpy()

    @property
    def pos(self) -> CoordSky:
        """Get a vector of the object sky positions in radians.

        Raises a :obj:`CachingError` if data is not loaded.

        Returns:
            :obj:`yaw.core.coordinates.CoordSky`
        """
        self.require_loaded()
        return CoordSky(self.ra, self.dec)

    @property
    def redshifts(self) -> NDArray[np.float_]:
        """Get the redshifts as array or ``None`` if not available.

        Raises a :obj:`CachingError` if data is not loaded."""
        self.require_loaded()
        if self.has_redshifts():
            return self._data["redshift"].to_numpy()
        else:
            return None

    @property
    def weights(self) -> NDArray[np.float_]:
        """Get the object weights as array or ``None`` if not available.

        Raises a :obj:`CachingError` if data is not loaded."""
        self.require_loaded()
        if self.has_weights():
            return self._data["weights"].to_numpy()
        else:
            return None

    @property
    def total(self) -> float:
        """Get the sum of weights or the number of objects if weights are not
        available.

        Available even if no data is loaded."""
        return self._total

    @property
    def center(self) -> CoordSky:
        """Get the patch centers in radians.

        Available even if no data is loaded.

        Returns:
            :obj:`yaw.core.coordinates.CoordSky`
        """
        return self._center.to_sky()

    @property
    def radius(self) -> DistSky:
        """Get the patch size in radians.

        Available even if no data is loaded.

        Returns:
            :obj:`yaw.core.coordinates.DistSky`
        """
        return self._radius

    def iter_bins(
        self, z_bins: NDArray[np.float_], allow_no_redshift: bool = False
    ) -> Iterator[tuple[Interval, PatchCatalog]]:
        """Iterate the patch in bins of redshift.

        Args:
            z_bins (:obj:`NDArray`):
                Edges of the redshift bins.
            allow_no_redshift (:obj:`bool`):
                If true and the data has no redshifts, the iterator yields the
                whole patch at each iteration step.

        Yields:
            (tuple): tuple containing:
                - **intv** (:obj:`pandas.Interval`): the selection for this bin.
                - **cat** (:obj:`PatchCatalog`): instance containing the data
                  for this bin.
        """
        if not allow_no_redshift and not self.has_redshifts():
            raise ValueError("no redshifts for iteration provdided")
        if allow_no_redshift:
            for intv in pd.IntervalIndex.from_breaks(z_bins, closed="left"):
                yield intv, self
        else:
            for intv, bin_data in self._data.groupby(pd.cut(self.redshifts, z_bins)):
                yield intv, PatchCatalog(
                    self.id,
                    bin_data,
                    degrees=False,
                    center=self._center,
                    radius=self._radius,
                )

    def get_tree(self, **kwargs) -> SphericalKDTree:
        """Build a :obj:`SphericalKDTree` from the patch data coordiantes."""
        tree = SphericalKDTree(self.pos, self.weights, **kwargs)
        tree._total = self.total  # no need to recompute this
        return tree


# Determine patch centers with k-means clustering. The implementation in
# treecorr is quite good, but might not be available. Implement a fallback using
# the scipy.cluster module.


def assign_patches(centers: Coordinate, position: Coordinate) -> NDArray[np.int_]:
    """Assign objects based on their coordinate to a list of points based on
    proximit."""
    patches, dist = vq.vq(position.to_3d().values, centers.to_3d().values)
    return patches


try:
    import treecorr

    def treecorr_patches(
        position: Coordinate, n_patches: int, **kwargs
    ) -> tuple[Coord3D, NDArray[np.int_]]:
        """Use the *k*-means clustering algorithm of :obj:`treecorr.Catalog` to
        generate spatial patches and assigning objects to those patches.
        """
        position = position.to_sky()
        cat = treecorr.Catalog(
            ra=position.ra,
            ra_units="radians",
            dec=position.dec,
            dec_units="radians",
            npatch=n_patches,
        )
        xyz = np.atleast_2d(cat.patch_centers)
        centers = Coord3D.from_array(xyz)
        if n_patches == 1:
            patches = np.zeros(len(position), dtype=np.int_)
        else:
            patches = assign_patches(centers=centers, position=position)
        del cat  # might not be necessary
        return centers, patches

    create_patches = treecorr_patches

except ImportError:

    def scipy_patches(
        position: Coordinate, n_patches: int, n_max: int = 500_000
    ) -> tuple[Coord3D, NDArray[np.int_]]:
        """Use the *k*-means clustering algorithm of :obj:`scipy.cluster` to
        generate spatial patches and assigning objects to those patches.
        """
        position = position.to_3d()
        subset = np.random.randint(0, len(position), size=min(n_max, len(position)))
        # place on unit sphere to avoid coordinate distortions
        centers, _ = vq.kmeans2(position[subset].values, n_patches, minit="points")
        centers = Coord3D.from_array(centers)
        patches = assign_patches(centers=centers, position=position)
        return centers, patches

    create_patches = scipy_patches



"""This module implements some functions for self-documentation and generation
of default configuration files.
"""

from __future__ import annotations

from collections.abc import Iterator, Mapping, Sequence
from dataclasses import MISSING, Field, asdict, dataclass, field, fields
from typing import TYPE_CHECKING, Any

import yaml

if TYPE_CHECKING:  # pragma: no cover
    from argparse import ArgumentParser

__all__ = ["Parameter", "get_doc_args", "populate_parser"]


@dataclass(frozen=True)
class Parameter(Mapping):
    """Class to implement automatic generation of parameter descriptions.

    This class provides the metadata needed to generate a new commandline
    argument for python's :mod:`argparse` module. Each constructor argument
    corresponds to one argument in the :func:`add_argument` method.

    Many examples can be found in the :mod:`yaw.config` module, e.g.:

    .. code-block:: python
        cosmology: TypeCosmology | str | None = field(
            default=DEFAULT.Configuration.cosmology,
            metadata=Parameter(
                type=str, choices=OPTIONS.cosmology,
                help="cosmological model used for distance calculations",
                default_text="(see astropy.cosmology, default: %(default)s)"))

    .. Note::
        The ``default_text`` parameter has to be provided separately and will be
        appended to the ``help``text when generating the full help text.
    """

    help: str
    type: type | None = field(default=None)
    nargs: str | int | None = field(default=None)
    choices: Sequence | None = field(default=None)
    required: bool = field(default=False)
    parser_id: str = field(default="default")
    default_text: str | None = field(default=None)
    metavar: str | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        if self.type not in (str, None) and not self.is_flag():
            try:
                _, rep, _ = str(self.type).split("'")
                metavar = f"<{rep}>"
            except ValueError:
                metavar = None
            object.__setattr__(self, "metavar", metavar)

    def __len__(self) -> int:
        return 5

    def __iter__(self) -> Iterator[str]:
        for cfield in fields(self):
            if cfield.init:
                yield f"yaw_{cfield.name}"

    def __getitem__(self, key: str) -> Any:
        return asdict(self)[key[4:]]

    @classmethod
    def from_field(cls, field: Field) -> Parameter:
        """Create a new instance from a :obj:`dataclassField`."""
        kwargs = {}
        for key, value in field.metadata.items():
            if key.startswith("yaw_"):
                kwargs[key[4:]] = value
        if len(kwargs) == 0:
            raise TypeError(
                f"cannot convert field with name '{field.name}' to Parameter"
            )
        return cls(**kwargs)

    def is_flag(self) -> bool:
        """Indicates if argument is a flag with ``action="store_true/false``."""
        return self.type is bool

    def get_kwargs(self) -> dict[str, Any]:
        """Build the list of keyword arguments for :meth:`add_argument`."""
        kwargs = asdict(self)
        kwargs.pop("parser_id")
        default = kwargs.pop("default_text")
        if default is not None:
            kwargs["help"] += " " + default
        return kwargs


def get_doc_args(
    dclass: object | type, indicate_opt: bool = True
) -> list[tuple[str, str | None]]:
    """Generate a section with default values for a YAML configuration file.

    Entries are added for those dataclass fields that contain a Parameter
    instance in the ``metadata`` field.
    """
    lines = []
    argfields = fields(dclass)
    if len(argfields) > 0:
        for cfield in argfields:
            try:  # omit parameter if not shipped with parameter information
                param = Parameter.from_field(cfield)
                # format the value as 'key: value'
                if cfield.default is not MISSING:
                    default = cfield.default
                    optional = True
                else:
                    default = None
                    optional = False
                value = yaml.dump({cfield.name.strip("_"): default}).strip()
                # format the optional comment
                comment = param.help
                if indicate_opt and optional:
                    comment = "(opt) " + comment
                if param.choices is not None:
                    comment += f" ({', '.join(param.choices)})"
                lines.append((value, comment))
            except TypeError:
                pass
    return lines


def populate_parser(
    dclass: object | type,
    default_parser: ArgumentParser,
    extra_parsers: Mapping[str, ArgumentParser] | None = None,
) -> None:
    """Populate a parser instance or argument group with arguments from a
    dataclass.

    Arguments are added for those dataclass fields that contain a Parameter
    instance in the ``metadata`` field.
    """
    for cfield in fields(dclass):
        try:
            parameter = Parameter.from_field(cfield)
        except TypeError:
            continue
        name = cfield.name.strip("_").replace("_", "-")

        if parameter.parser_id == "default":
            parser = default_parser
        else:
            parser = extra_parsers[parameter.parser_id]

        if parameter.is_flag():
            if cfield.default is True:
                parser.add_argument(
                    f"--no-{name}",
                    dest=cfield.name,
                    action="store_false",
                    help=parameter.help,
                )
            else:
                parser.add_argument(
                    f"--{name}", action="store_true", help=parameter.help
                )

        else:
            kwargs = parameter.get_kwargs()
            if cfield.default is not MISSING:
                kwargs["default"] = cfield.default
            parser.add_argument(f"--{name}", **kwargs)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import numpy as np

from yaw.config import OPTIONS
from yaw.config import default as DEFAULT
from yaw.config.utils import ConfigError
from yaw.core.abc import DictRepresentation
from yaw.core.cosmology import BinFactory, TypeCosmology
from yaw.core.docs import Parameter
from yaw.core.math import array_equal

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

__all__ = ["AutoBinningConfig", "ManualBinningConfig", "make_binning_config"]


logger = logging.getLogger(__name__)


class BaseBinningConfig(DictRepresentation):
    """Base class for redshift binning configuration."""

    zbins: NDArray[np.float_]
    """Edges of redshift bins."""
    method: str
    """Method used to create redshift binning, ``manual`` or either of
    :obj:`~yaw.config.options.Options.binning`."""

    def __repr__(self) -> str:
        name = self.__class__.__name__
        zbin_num = self.zbin_num
        z = f"{self.zmin:.3f}...{self.zmax:.3f}"
        method = self.method
        return f"{name}({zbin_num=}, {z=}, {method=})"


@dataclass(frozen=True, repr=False)
class ManualBinningConfig(BaseBinningConfig):
    """Configuration that specifies a manual redshift binning.

    Args:
        zbins (:obj:`NDArray`):
            Edges of redshift bins, must increase monotonically.
    """

    zbins: NDArray[np.float_] = field(
        metadata=Parameter(
            type=float,
            nargs="*",
            help="list of custom redshift bin edges, if provided, other "
            "binning parameters are omitted, method is set to 'manual'",
        )
    )

    def __post_init__(self) -> None:
        if len(self.zbins) < 2:
            raise ConfigError("'zbins' must have at least two edges")
        BinFactory.check(self.zbins)
        object.__setattr__(self, "zbins", np.asarray(self.zbins))

    def __eq__(self, other: ManualBinningConfig) -> bool:
        if not array_equal(self.zbins, other.zbins):
            return False
        return True

    @property
    def method(self) -> str:
        """Method used to create redshift binning, always ``manual``."""
        return "manual"

    @property
    def zmin(self) -> float:
        """Lowest redshift bin edge."""
        return float(self.zbins[0])

    @property
    def zmax(self) -> float:
        """Highest redshift bin edge."""
        return float(self.zbins[-1])

    @property
    def zbin_num(self) -> int:
        """Number of redshift bins."""
        return len(self.zbins) - 1

    @classmethod
    def from_dict(cls, the_dict: dict[str, Any], **kwargs) -> ManualBinningConfig:
        return cls(np.asarray(the_dict["zbins"]))

    def to_dict(self) -> dict[str, Any]:
        return dict(method=self.method, zbins=self.zbins.tolist())


@dataclass(frozen=True, repr=False)
class AutoBinningConfig(BaseBinningConfig):
    """Configuration that generates a redshift binning.

    To generate a redshift binning use the :meth:`generate` method.

    Args:
        zbins (:obj:`NDArray`):
            Edges of redshift bins, must increase monotonically.
        method (:obj:`str`):
            Method used to create redshift binning, either of
            :obj:`~yaw.config.options.Options.binning`.
    """

    zbins: NDArray[np.float_]
    method: str = field(
        default=DEFAULT.Configuration.binning.method,
        metadata=Parameter(
            choices=OPTIONS.binning,
            help="redshift binning method, 'logspace' means equal size in log(1+z)",
            default_text="(default: %(default)s)",
        ),
    )
    """Method used to create redshift binning, see
    :obj:`~yaw.config.options.Options.binning`."""
    zmin: float = field(
        init=False,
        metadata=Parameter(
            type=float,
            help="lower redshift limit",
            default_text="(default: %(default)s)",
        ),
    )
    """Lowest redshift bin edge."""
    zmax: float = field(
        init=False,
        metadata=Parameter(
            type=float,
            help="upper redshift limit",
            default_text="(default: %(default)s)",
        ),
    )
    """Highest redshift bin edge."""
    zbin_num: int = field(
        default=DEFAULT.AutoBinning.zbin_num,
        init=False,
        metadata=Parameter(
            type=int,
            help="number of redshift bins",
            default_text="(default: %(default)s)",
        ),
    )
    """Number of redshift bins."""

    def __post_init__(self) -> None:
        object.__setattr__(self, "zmin", float(self.zbins[0]))
        object.__setattr__(self, "zmax", float(self.zbins[-1]))
        object.__setattr__(self, "zbin_num", len(self.zbins) - 1)

    @classmethod
    def generate(
        cls,
        zmin: float,
        zmax: float,
        zbin_num: int = DEFAULT.AutoBinning.zbin_num,
        method: str = DEFAULT.AutoBinning.method,
        cosmology: TypeCosmology | None = None,
    ) -> AutoBinningConfig:
        """Generate a new redshift binning configuration.

        Generates a specified number of bins between a lower and upper redshift
        limit. The spacing of the bins depends the generation method, the
        default is a linear spacing.

        Args:
            zmin (:obj:`float`):
                Minimum redshift, lowest redshift edge.
            zmax (:obj:`float`):
                Maximum redshift, highest redshift edge.
            zbin_num (:obj:`int`, optional):
                Number of redshift bins to generate.
            method (:obj:`str`, optional):
                Method used to create redshift binning, for a list of valid
                options and their description see
                :obj:`~yaw.config.options.Options.binning`.
            cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`~yaw.core.cosmology.CustomCosmology`, optional):
                Cosmological model used for distance calculations. For a custom
                model, refer to :mod:`yaw.core.cosmology`.

        Returns:
            :obj:`AutoBinningConfig`
        """
        zbins = BinFactory(zmin, zmax, zbin_num, cosmology).get(method)
        return cls(zbins, method)

    def __eq__(self, other: ManualBinningConfig) -> bool:
        if not array_equal(self.zbins, other.zbins):
            return False
        if self.method != other.method:
            return False
        return True

    @classmethod
    def from_dict(
        cls, the_dict: dict[str, Any], cosmology: TypeCosmology | None = None
    ) -> AutoBinningConfig:
        """Create a class instance from a dictionary representation of the
        minimally required data.

        Args:
            the_dict (:obj:`dict`):
                Dictionary containing the data.
            cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`~yaw.core.cosmology.CustomCosmology`, optional):
                Cosmological model used for distance calculations. For a custom
                model, refer to :mod:`yaw.core.cosmology`.
        """
        return cls.generate(**the_dict, cosmology=cosmology)

    def to_dict(self) -> dict[str, Any]:
        return dict(
            zmin=self.zmin, zmax=self.zmax, zbin_num=self.zbin_num, method=self.method
        )


def warn_binning_args_ignored(
    zmin: bool | float | None, zmax: bool | float | None, zbin_num: bool | int | None
) -> None:
    """Issue a warning that values are ignored, if any of the arguments is
    set."""
    # NOTE: NotSet is also False
    if zmin or zmax or zbin_num:
        logger.warn("'zmin', 'zmax', 'nbins' are ignored if 'zbins' is provided")


def make_binning_config(
    cosmology: TypeCosmology | str | None,
    zmin: float | None = None,
    zmax: float | None = None,
    zbin_num: int | None = None,
    method: str | None = None,
    zbins: NDArray[np.float_] | None = None,
) -> ManualBinningConfig | AutoBinningConfig:
    """
    Helper function to construct a binning configuration.

    The ``cosmology`` argument is always required. If redshift bins (``zbins``)
    are provided, a :obj:`ManualBinningConfig` is returned, otherwise an
    :obj:`AutoBinningConfig`. Issues a warning if any argument is set but
    ignored by the returned configuration instance.
    """
    auto_args_set = (zmin is not None, zmax is not None, zbin_num is not None)
    if zbins is None and not all(auto_args_set):
        raise ConfigError("either 'zbins' or 'zmin', 'zmax', 'zbin_num' are required")
    elif all(auto_args_set):
        return AutoBinningConfig.generate(
            zmin=zmin, zmax=zmax, zbin_num=zbin_num, method=method, cosmology=cosmology
        )
    else:
        warn_binning_args_ignored(*auto_args_set)
        return ManualBinningConfig(zbins)



from __future__ import annotations

import logging
from abc import abstractclassmethod, abstractmethod, abstractproperty
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, TypeVar

import astropandas as apd
import numpy as np
from numpy.typing import NDArray

from yaw.core.coordinates import Coordinate, CoordSky, DistSky
from yaw.core.utils import long_num_format

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame

    from yaw.catalogs import PatchLinkage
    from yaw.config import Configuration, ResamplingConfig
    from yaw.correlation.paircounts import NormalisedCounts
    from yaw.redshifts import HistData

__all__ = ["BaseCatalog"]


_Tcat = TypeVar("_Tcat", bound="BaseCatalog")


class BackendError(Exception):
    pass


class BaseCatalog:
    """The data catalog base class.

    Every new backend must implement a catalog class based on this abstract base
    class. On import this subclass is automatically registered and can be
    instantiated using the factory class :class:`yaw.NewCatalog`.

    .. Note::

        Base classes must follow the ``[Backendname]Catalog`` naming convention.
        The new backend is then registered with name ``backendname`` (lower
        case).
    """

    _logger = logging.getLogger("yaw.catalog")
    _backends = dict()

    def __init_subclass__(cls, **kwargs):
        """Handles the backend subclass registration."""
        super().__init_subclass__(**kwargs)
        if not cls.__name__.endswith("Catalog"):
            raise BackendError(
                "subclasses of 'BaseCatalog' must follow naming convention "
                "'[Backend name]Catalog for registration (e.g. ScipyCatalog "
                "-> 'scipy')"
            )
        backend = cls.__name__.strip("Catalog").lower()
        cls._backends[backend] = cls

    @abstractmethod
    def __init__(
        self,
        data: DataFrame,
        ra_name: str,
        dec_name: str,
        *,
        patch_name: str | None = None,
        patch_centers: BaseCatalog | Coordinate | None = None,
        n_patches: int | None = None,
        redshift_name: str | None = None,
        weight_name: str | None = None,
        cache_directory: str | None = None,
        progress: bool = False,
    ) -> None:
        """Build a catalogue from in-memory data.

        Catalogs should be instantiated through the factory class, see
        :meth:`yaw.catalogs.NewCatalog.from_dataframe`."""
        pass

    @classmethod
    def from_file(
        cls,
        filepath: str,
        patches: str | int | BaseCatalog | Coordinate,
        ra: str,
        dec: str,
        *,
        redshift: str | None = None,
        weight: str | None = None,
        sparse: int | None = None,
        cache_directory: str | None = None,
        file_ext: str | None = None,
        progress: bool = False,
        **kwargs,
    ) -> BaseCatalog:
        """Build a catalogue from data file.

        Catalogs should be instantiated through the factory class, see
        :meth:`yaw.catalogs.NewCatalog.from_file`."""
        columns = [c for c in [ra, dec, redshift, weight] if c is not None]
        if isinstance(patches, str):
            columns.append(patches)
            patch_kwarg = dict(patch_name=patches)
        elif isinstance(patches, int):
            patch_kwarg = dict(n_patches=patches)
        elif isinstance(patches, Coordinate):
            patch_kwarg = dict(patch_centers=patches)
        elif isinstance(patches, BaseCatalog):
            patch_kwarg = dict(patch_centers=patches.centers)
        else:
            raise TypeError(
                "'patches' must be either of type 'str' (col. name), 'int' "
                "(number of patches), or 'Catalog' or 'Coordinate' (specify "
                "centers)"
            )

        cls._logger.info(f"reading catalog file '{filepath}'")
        data = apd.read_auto(filepath, columns=columns, ext=file_ext, **kwargs)
        if sparse is not None:
            cls._logger.debug(f"sparse sampling data {sparse}x")
            data = data[::sparse]
        return cls(
            data,
            ra,
            dec,
            **patch_kwarg,
            redshift_name=redshift,
            weight_name=weight,
            cache_directory=cache_directory,
            progress=progress,
        )

    @abstractclassmethod
    def from_cache(cls, cache_directory: str, progress: bool = False) -> BaseCatalog:
        """Restore the catalogue from its cache directory.

        Catalogs should be instantiated through the factory class, see
        :meth:`yaw.catalogs.NewCatalog.from_cache`."""
        cls._logger.info(f"restoring from cache directory '{cache_directory}'")

    def __repr__(self) -> str:
        name = self.__class__.__name__
        args = dict(
            loaded=self.is_loaded(),
            nobjects=len(self),
            npatches=self.n_patches,
            redshifts=self.has_redshifts(),
        )
        arg_str = ", ".join(f"{k}={v}" for k, v in args.items())
        return f"{name}({arg_str})"

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> Any:
        pass

    @abstractproperty
    def ids(self) -> list[int]:
        """Return a list of unique patch indices in the catalog."""
        pass

    @abstractmethod
    def n_patches(self) -> int:
        """The number of spatial patches of this catalogue."""
        pass

    @abstractmethod
    def __iter__(self) -> Iterator:
        pass

    @abstractmethod
    def is_loaded(self) -> bool:
        """Indicates whether the catalog data is loaded.

        Always ``True`` if no cache is used. If the catalog is unloaded, data
        will be read from cache every time data is accessed."""
        pass

    @abstractmethod
    def load(self) -> None:
        """Permanently load data from cache into memory.

        Raises a :obj:`~yaw.catalogs.scipy.patches.CachingError` if no cache
        is configured.
        """
        self._logger.debug("bulk loading catalog")

    @abstractmethod
    def unload(self) -> None:
        """Unload data from memory if a disk cache is provided."""
        self._logger.debug("bulk unloading catalog")

    @abstractmethod
    def has_redshifts(self) -> bool:
        """Indicates whether the :meth:`redshifts` attribute holds data."""
        pass

    @abstractmethod
    def has_weights(self) -> bool:
        """Indicates whether the :meth:`weights` attribute holds data."""
        pass

    @property
    def pos(self) -> CoordSky:
        """Get a vector of the object sky positions in radians.

        Returns:
            :obj:`yaw.core.coordinates.CoordSky`
        """
        return CoordSky(self.ra, self.dec)

    @abstractproperty
    def ra(self) -> NDArray[np.float_]:
        """Get an array of the right ascension values in radians."""
        pass

    @abstractproperty
    def dec(self) -> NDArray[np.float_]:
        """Get an array of the declination values in radians."""
        pass

    @abstractproperty
    def redshifts(self) -> NDArray[np.float_] | None:
        """Get the redshifts as array or ``None`` if not available."""
        pass

    @abstractproperty
    def weights(self) -> NDArray[np.float_]:
        """Get the object weights as array or ``None`` if not available."""
        pass

    @abstractproperty
    def patch(self) -> NDArray[np.int_]:
        """Get the patch indices of each object as array."""
        pass

    @abstractmethod
    def get_min_redshift(self) -> float:
        """Get the minimum redshift or ``None`` if not available."""
        pass

    @abstractmethod
    def get_max_redshift(self) -> float:
        """Get the maximum redshift or ``None`` if not available."""
        pass

    @abstractproperty
    def total(self) -> float:
        """Get the sum of weights or the number of objects if weights are not
        available."""

    @abstractmethod
    def get_totals(self) -> NDArray[np.float_]:
        """Get an array of the sum of weights or number of objects in each
        patch."""

    @abstractproperty
    def centers(self) -> CoordSky:
        """Get a vector of sky coordinates of the patch centers in radians.

        Returns:
            :obj:`yaw.core.coordinates.CoordSky`
        """
        pass

    @abstractproperty
    def radii(self) -> DistSky:
        """Get a vector of angular separations in radians that describe the
        patch sizes.

        The radius of the patch is defined as the maximum angular distance of
        any object from the patch center.

        Returns:
            :obj:`yaw.core.coordinates.DistSky`
        """
        pass

    @abstractmethod
    def correlate(
        self,
        config: Configuration,
        binned: bool,
        other: _Tcat = None,
        linkage: PatchLinkage | None = None,
        progress: bool = False,
    ) -> NormalisedCounts | dict[str, NormalisedCounts]:
        """Count pairs between objects at a given separation and in bins of
        redshift.

        If another catalog instance is passed to ``other``, then pairs are
        formed between these catalogues (cross), otherwise pairs are formed with
        the catalog (auto). Pairs are counted in bins of redshift, as defined in
        the configuration object (``config``). Pairs are only considered within
        fixed angular scales that are computed from the physical scales in the
        configuration and the mid of the current redshift bin.

        Args:
            config (:obj:`yaw.Configuration`):
                Configuration object that defines measurement scales, redshift
                binning, cosmological model, and various backend specific
                parameters.
            binned (:obj:`bool`):
                Whether to apply the redshift binning to the second catalogue
                (see ``other``).
            other (Catalog instance, optional):
                Second catalog instance used for cross-catalogue pair counting.
                Catalogue must use the same backend.
            linkage (:obj:`~yaw.catalogs.linkage.PatchLinkage`, optional):
                Linkage object that defines with patches must be correlated for
                a given scales and which patch combinations can be skipped. Can
                be used for the ``scipy`` backend to count pairs consistently
                between multiple catalogue instances.
            progress (:obj:`bool`):
                Show a progress indication, depends on backend.

        There are three different modes of operation that are determined by the
        combination of the ``binned`` and ``other`` parameters:

        1. If no second catalogue is provided, pairs are counted within the
           catalogue while applying the redshift binning.
        2. If a second catalogue is provided and ``binned=True``, pairs are
           counted between the catalogues and the binning is applied to both
           cataluges.
        3. If a second catalogue is provided and ``binned=False``, the redshift
           binning is not applied to the second catalogue, otherwise above.

        The catalogue from the calling instance of :meth:`correlate` has always
        redshift binning applied.
        """
        n1 = long_num_format(len(self))
        n2 = long_num_format(len(self) if other is None else len(other))
        self._logger.debug(
            f"correlating with {'' if binned else 'un'}binned catalog "
            f"({n1}x{n2}) in {config.binning.zbin_num} redshift bins"
        )

    @abstractmethod
    def true_redshifts(
        self,
        config: Configuration,
        sampling_config: ResamplingConfig | None = None,
        progress: bool = False,
    ) -> HistData:
        """
        Compute a histogram of the object redshifts from the binning defined in
        the provided configuration.

        Args:
            config (:obj:`~yaw.config.Configuration`):
                Defines the bin edges used for the histogram.
            sampling_config (:obj:`~yaw.config.ResamplingConfig`, optional):
                Specifies the spatial resampling for error estimates.
            progress (:obj:`bool`):
                Show a progress bar.

        Returns:
            HistData:
                Object holding the redshift histogram
        """
        self._logger.info("computing true redshift distribution")



"""This module implements functions and wrappers for handling logs and warnings.
"""

from __future__ import annotations

import logging
import warnings
from datetime import timedelta
from timeit import default_timer
from typing import Callable

__all__ = ["LogCustomWarning", "TimedLog"]


class LogCustomWarning:
    """Context wrapper that temporarily redirects warnings to a logger."""

    def __init__(
        self,
        logger: logging.Logger,
        alt_message: str | None = None,
        ignore: bool = True,
    ):
        """Instead of showing the warning through the :func:`warnings.warn`
        machinery, write the message as warning to the provided logger.

        Args:
            logger (:obj:`logging.Logger`):
                The logger instance to which the warning is redirected.
            alt_message (:obj:`str`, optional):
                Replace the original message text with this value instead.
            ignore (:obj:`bool`, optional):
                Do not show warning with :func:`warnings.warn` (the default).
        """
        self._logger = logger
        self._message = alt_message
        self._ignore = ignore

    def _process_warning(self, message, category, filename, lineno, *args):
        if not self._ignore:
            self._old_showwarning(message, category, filename, lineno, *args)
        if self._message is not None:
            message = self._message
        else:
            message = f"{category.__name__}: {message}"
        self._logger.warn(message)

    def __enter__(self) -> TimedLog:
        self._old_showwarning = warnings.showwarning
        warnings.showwarning = self._process_warning
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        warnings.showwarning = self._old_showwarning


class TimedLog:
    """Context wrapper that measures the elapsed time and emits a log message on
    exit.

    Emits a log in the format ``{message} - done {elapsed time}``.

    Args:
        logging_callback (Callable):
            Function that processes the log message on context wrapper exit.
        msg (:obj:`str`, optional):
            The log message body.
    """

    def __init__(self, logging_callback: Callable, msg: str | None = None) -> None:
        self.callback = logging_callback
        self.msg = msg

    def __enter__(self) -> TimedLog:
        self.t = default_timer()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        delta = default_timer() - self.t
        time = str(timedelta(seconds=round(delta)))
        self.callback(f"{self.msg} - done {time}")
